# Authors:
#   Pavel Zuna <pzuna@redhat.com>
#   Adam Young <ayoung@redhat.com>
#   Endi S. Dewata <edewata@redhat.com>
#
# Copyright (c) 2010  Red Hat
# See file 'copying' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ipalib import Command
from ipalib import Str
from ipalib.frontend import Local
from ipalib.output import Output
from ipalib.text import _
from ipalib.util import json_serialize
from ipalib.plugable import Registry

__doc__ = _("""
Plugins not accessible directly through the CLI, commands used internally
""")

register = Registry()

@register()
class json_metadata(Command):
    __doc__ = _('Export plugin meta-data for the webUI.')
    NO_CLI = True


    takes_args = (
        Str('objname?',
            doc=_('Name of object to export'),
        ),
        Str('methodname?',
            doc=_('Name of method to export'),
        ),
    )

    takes_options = (
        Str('object?',
            doc=_('Name of object to export'),
        ),
        Str('method?',
            doc=_('Name of method to export'),
        ),
        Str('command?',
            doc=_('Name of command to export'),
        ),
    )

    has_output = (
        Output('objects', dict, doc=_('Dict of JSON encoded IPA Objects')),
        Output('methods', dict, doc=_('Dict of JSON encoded IPA Methods')),
        Output('commands', dict, doc=_('Dict of JSON encoded IPA Commands')),
    )

    def execute(self, objname=None, methodname=None, **options):
        objects = dict()
        methods = dict()
        commands = dict()

        empty = True

        try:
            if not objname:
                objname = options['object']
            if objname in self.api.Object:
                o = self.api.Object[objname]
                objects = dict([(o.name, json_serialize(o))])
            elif objname == "all":
                objects = dict(
                    (o.name, json_serialize(o)) for o in self.api.Object()
                    if o is self.api.Object[o.name]
                )
            empty = False
        except KeyError:
            pass

        try:
            if not methodname:
                methodname = options['method']
            if (methodname in self.api.Method and
                    not isinstance(self.api.Method[methodname], Local)):
                m = self.api.Method[methodname]
                methods = dict([(m.name, json_serialize(m))])
            elif methodname == "all":
                methods = dict(
                    (m.name, json_serialize(m)) for m in self.api.Method()
                    if (m is self.api.Method[m.name] and
                        not isinstance(m, Local))
                )
            empty = False
        except KeyError:
            pass

        try:
            cmdname = options['command']
            if (cmdname in self.api.Command and
                    not isinstance(self.api.Command[cmdname], Local)):
                c = self.api.Command[cmdname]
                commands = dict([(c.name, json_serialize(c))])
            elif cmdname == "all":
                commands = dict(
                    (c.name, json_serialize(c)) for c in self.api.Command()
                    if (c is self.api.Command[c.name] and
                        not isinstance(c, Local))
                )
            empty = False
        except KeyError:
            pass

        if empty:
            objects = dict(
                (o.name, json_serialize(o)) for o in self.api.Object()
                if o is self.api.Object[o.name]
            )
            methods = dict(
                (m.name, json_serialize(m)) for m in self.api.Method()
                if (m is self.api.Method[m.name] and
                    not isinstance(m, Local))
            )
            commands = dict(
                (c.name, json_serialize(c)) for c in self.api.Command()
                if (c is self.api.Command[c.name] and
                    not isinstance(c, Local))
            )

        retval = dict([
            ("objects", objects),
            ("methods", methods),
            ("commands", commands),
        ])

        return retval


@register()
class i18n_messages(Command):
    __doc__ = _('Internationalization messages')
    NO_CLI = True

    messages = {
        "ajax": {
            "401": {
                "message": _("Your session has expired. Please log in again."),
            },
        },
        "actions": {
            "apply": _("Apply"),
            "automember_rebuild": _("Rebuild auto membership"),
            "automember_rebuild_confirm": _("Are you sure you want to rebuild auto membership?"),
            "automember_rebuild_success": _("Automember rebuild membership task completed"),
            "confirm": _("Are you sure you want to proceed with the action?"),
            "delete_confirm": _("Are you sure you want to delete ${object}?"),
            "disable_confirm": _("Are you sure you want to disable ${object}?"),
            "enable_confirm": _("Are you sure you want to enable ${object}?"),
            "title": _("Actions"),
        },
        "association": {
            "add_title_default": _("Add"),
            "added": _("${count} item(s) added"),
            "direct_membership": _("Direct Membership"),
            "filter_placeholder": _("Filter available ${other_entity}"),
            "indirect_membership": _("Indirect Membership"),
            "no_entries": _("No entries."),
            "paging": _("Showing ${start} to ${end} of ${total} entries."),
            "remove_title_default": _("Remove"),
            "removed": _("${count} item(s) removed"),
            "show_results": _("Show Results"),
        },
        "authtype": {
            "auth_indicators": _("Authentication indicators"),
            "auth_indicator": _("Authentication indicator"),
            "config_tooltip": _("<p>Implicit method (password) will be used if no method is chosen.</p><p><strong>Password + Two-factor:</strong> LDAP and Kerberos allow authentication with either one of the authentication types but Kerberos uses pre-authentication method which requires to use armor ccache.</p><p><strong>RADIUS with another type:</strong> Kerberos always use RADIUS, but LDAP never does. LDAP only recognize the password and two-factor authentication options.</p>"),
            "custom_auth_ind_title": _("Add Custom Authentication Indicator"),
            "otp": _("OTP"),
            "type_otp": _("Two factor authentication (password + OTP)"),
            "type_password": _("Password"),
            "type_radius": _("RADIUS"),
            "type_disabled": _("Disable per-user override"),
            "user_tooltip": _("<p>Per-user setting, overwrites the global setting if any option is checked.</p><p><strong>Password + Two-factor:</strong> LDAP and Kerberos allow authentication with either one of the authentication types but Kerberos uses pre-authentication method which requires to use armor ccache.</p><p><strong>RADIUS with another type:</strong> Kerberos always use RADIUS, but LDAP never does. LDAP only recognize the password and two-factor authentication options.</p>"),
        },
        "buttons": {
            "about": _("About"),
            "activate": _("Activate"),
            "add": _("Add"),
            "add_and_add_another": _("Add and Add Another"),
            "add_and_close": _("Add and Close"),
            "add_and_edit": _("Add and Edit"),
            "add_many": _("Add Many"),
            "apply": _("Apply"),
            "back": _("Back"),
            "cancel": _("Cancel"),
            "clear": _("Clear"),
            "clear_title": _("Clear all fields on the page."),
            "close": _("Close"),
            "disable": _("Disable"),
            "download": _("Download"),
            "download_title": _("Download certificate as PEM formatted file."),
            "edit": _("Edit"),
            "enable": _("Enable"),
            "filter": _("Filter"),
            "find": _("Find"),
            "get": _("Get"),
            "hide": _("Hide"),
            "issue": _("Issue"),
            "match": _("Match"),
            "match_title": _("Match users according to certificate."),
            "migrate": _("Migrate"),
            "ok": _("OK"),
            "refresh": _("Refresh"),
            "refresh_title": _("Reload current settings from the server."),
            "remove": _("Delete"),
            "remove_hold": _("Remove hold"),
            "reset": _("Reset"),
            "reset_password": _("Reset Password"),
            "reset_password_and_login": _("Reset Password and Log in"),
            "restore": _("Restore"),
            "retry": _("Retry"),
            "revert": _("Revert"),
            "revert_title": ("Undo all unsaved changes."),
            "revoke": _("Revoke"),
            "save": _("Save"),
            "set": _("Set"),
            "show": _("Show"),
            "stage": _("Stage"),
            "unapply": ("Un-apply"),
            "update": _("Update"),
            "view": _("View"),
        },
        "customization": {
            "customization": _("Customization"),
            "table_pagination": _("Pagination Size"),
        },
        "details": {
            "collapse_all": _("Collapse All"),
            "expand_all": _("Expand All"),
            "general": _("General"),
            "identity": _("Identity Settings"),
            "settings": _("${entity} ${primary_key} Settings"),
            "to_top": _("Back to Top"),
            "updated": _("${entity} ${primary_key} updated"),
        },
        "dialogs": {
            "add_confirmation": _("${entity} successfully added"),
            "add_custom_value": _("Add custom value"),
            "add_title_default": _("Add"),
            "available": _("Available"),
            "batch_error_message": _("Some operations failed."),
            "batch_error_title": _("Operations Error"),
            "confirmation": _("Confirmation"),
            "custom_value": _("Custom value"),
            "dirty_message": _("This page has unsaved changes. Please save or revert."),
            "dirty_title": _("Unsaved Changes"),
            "edit_title": _("Edit ${entity}"),
            "hide_details": _("Hide details"),
            "about_title": _("About"),
            "about_message": _("${product}, version: ${version}"),
            "prospective": _("Prospective"),
            "redirection": _("Redirection"),
            "remove_empty": _("Select entries to be removed."),
            "remove_title_default": _("Remove"),
            "result": _("Result"),
            "show_details": _("Show details"),
            "success": _("Success"),
            "validation_title": _("Validation error"),
            "validation_message": _("Input form contains invalid or missing values."),
        },
        "error_report": {
            "options": _("Please try the following options:"),
            "problem_persists": _("If the problem persists please contact the system administrator."),
            "refresh": _("Refresh the page."),
            "reload": _("Reload the browser."),
            "main_page": _("Return to the main page and retry the operation"),
            "title": _("An error has occurred (${error})"),
        },
        "errors": {
            "error": _("Error"),
            "http_error": _("HTTP Error"),
            "internal_error": _("Internal Error"),
            "ipa_error": _("IPA Error"),
            "no_response": _("No response"),
            "unknown_error": _("Unknown Error"),
            "url": _("URL"),
        },
        "facet_groups": {
            "managedby": _("${primary_key} is managed by:"),
            "member": _("${primary_key} members:"),
            "memberof": _("${primary_key} is a member of:"),
        },
        "facets": {
            "details": _("Settings"),
            "search": _("Search"),
        },
        "false": _("False"),
        "keytab": {
            "add_groups_create": _(
                "Allow user groups to create keytab of '${primary_key}'"
            ),
            "add_groups_retrieve": _(
                "Allow user groups to retrieve keytab of '${primary_key}'"
            ),
            "add_hostgroups_create": _(
                "Allow host groups to create keytab of '${primary_key}'"
            ),
            "add_hostgroups_retrieve": _(
                "Allow host groups to retrieve keytab of '${primary_key}'"
            ),
            "add_hosts_create": _(
                "Allow hosts to create keytab of '${primary_key}'"
            ),
            "add_hosts_retrieve": _(
                "Allow hosts to retrieve keytab of '${primary_key}'"
            ),
            "add_users_create": _(
                "Allow users to create keytab of '${primary_key}'"
            ),
            "add_users_retrieve": _(
                "Allow users to retrieve keytab of '${primary_key}'"
            ),
            "allowed_to_create": _("Allowed to create keytab"),
            "allowed_to_retrieve": _("Allowed to retrieve keytab"),
            "remove_groups_create": _(
                "Disallow user groups to create keytab of '${primary_key}'"
            ),
            "remove_groups_retrieve": _(
                "Disallow user groups to retrieve keytab of '${primary_key}'"
            ),
            "remove_hostgroups_create": _(
                "Disallow host groups to create keytab of '${primary_key}'"
            ),
            "remove_hostgroups_retrieve": _(
                "Disallow host groups to retrieve keytab of '${primary_key}'"
            ),
            "remove_hosts_create": _(
                "Disallow hosts to create keytab of '${primary_key}'"
            ),
            "remove_hosts_retrieve": _(
                "Disallow hosts to retrieve keytab of '${primary_key}'"
            ),
            "remove_users_create": _(
                "Disallow users to create keytab of '${primary_key}'"
            ),
            "remove_users_retrieve": _(
                "Disallow users to retrieve keytab of '${primary_key}'"
            ),
        },
        "krbaliases": {
            "adder_title": _("Add Kerberos Principal Alias"),
            "add_krbal_label": _("New kerberos principal alias"),
            "remove_title": _("Remove Kerberos Alias"),
            "remove_message": _("Do you want to remove kerberos alias ${alias}?"),
        },
        "krbauthzdata": {
            "inherited": _("Inherited from server configuration"),
            "mspac": _("MS-PAC"),
            "override": _("Override inherited settings"),
            "pad": _("PAD"),
        },
        "login": {
            "authenticating": _("Authenticating"),
            "cert_auth_failed": _(
                "Authentication with personal certificate failed"),
            "cert_msg": _(
                "<i class=\"fa fa-info-circle\"></i> To log in with "
                "<strong>certificate</strong>, please make sure you have "
                "valid personal certificate. "
            ),
            "continue_msg": _("Continue to next page"),
            "form_auth": _(
                "<i class=\"fa fa-info-circle\"></i> To log in with "
                "<strong>username and password</strong>, enter them in the "
                "corresponding fields, then click 'Log in'."),
            "form_auth_failed": _("Login failed due to an unknown reason"),
            "header": _("Logged In As"),
            "krb_auth_failed": _("Authentication with Kerberos failed"),
            "krb_auth_msg": _(
                "<i class=\"fa fa-info-circle\"></i> To log in with "
                "<strong>Kerberos</strong>, please make sure you have valid "
                "tickets (obtainable via kinit) and <a href='${protocol}//"
                "${host}/ipa/config/ssbrowser.html'>configured</a> the browser"
                " correctly, then click 'Log in'."),
            "loading": _("Loading"),
            "krbprincipal_expired": _(
                "Kerberos Principal you entered is expired"),
            "loading_md": _("Loading data"),
            "login": _("Log in"),
            "login_certificate": _("Log In Using Certificate"),
            "login_certificate_desc": _("Log in using personal certificate"),
            "logout": _("Log out"),
            "logout_error": _("Log out error"),
            "password": _("Password"),
            "password_and_otp": _("Password or Password+One-Time-Password"),
            "redirect_msg": _("You will be redirected in ${count}s"),
            "sync_otp_token": _("Sync OTP Token"),
            "synchronizing": _("Synchronizing"),
            "username": _("Username"),
            "user_locked": _("The user account you entered is locked"),
        },
        "measurement_units": {
            "number_of_passwords": _("number of passwords"),
            "seconds": _("seconds"),
        },
        "migration": {
            "migrating": _("Migrating"),
            "migration_error_msg": _(
                "There was a problem with your request. Please, try again "
                "later."),
            "migration_failure_msg": _(
                "Password migration was not successful"),
            "migration_info_msg": _(
                "<h1>Password Migration</h1><p>If you have been sent here by "
                "your administrator, your personal information is being "
                "migrated to a new identity management solution (IPA).</p><p>"
                "Please, enter your credentials in the form to complete the "
                "process. Upon successful login your kerberos account will be "
                "activated.</p>"),
            "migration_invalid_password": _(
                "The password or username you entered is incorrect"),
            "migration_success": _("Password migration was successful"),
        },
        "objects": {
            "aci": {
                "attribute": _("Attribute"),
            },
            "acidelegation": {
                "add": _("Add delegation"),
                "remove": _("Remove delegations"),
            },
            "acipermission": {
                "add": _("Add permission"),
                "add_privileges": _(
                    "Add privileges into permission '${primary_key}'"
                ),
                "remove": _("Remove permissions"),
                "remove_privileges": _(
                    "Remove privileges from permission '${primary_key}'"
                ),
            },
            "aciprivilege": {
                "add": _("Add privilege"),
                "add_into_permissions": _(
                    "Add privilege '${primary_key}' into permissions"
                ),
                "add_roles": _(
                    "Add roles into privilege '${primary_key}'"
                ),
                "remove": _("Remove privileges"),
                "remove_from_permissions": _(
                    "Remove privilege '${primary_key}' from permissions"
                ),
                "remove_roles": _(
                    "Remove roles from privilege '${primary_key}'"
                ),
            },
            "acirole": {
                "identity": _("Role Settings"),
                "add": _("Add role"),
                "add_groups": _(
                    "Add user groups into role '${primary_key}'"
                ),
                "add_hosts": _(
                    "Add hosts into role '${primary_key}'"
                ),
                "add_hostgroups": _(
                    "Add host groups into role '${primary_key}'"
                ),
                "add_into_privileges": _(
                    "Add role '${primary_key}' into privileges"
                ),
                "add_services": _(
                    "Add services into role '${primary_key}'"
                ),
                "add_users": _(
                    "Add users into role '${primary_key}'"
                ),
                "remove": _("Remove roles"),
                "remove_from_privileges": _(
                    "Remove role '${primary_key}' from privileges"
                ),
                "remove_groups": _(
                    "Remove user groups from role '${primary_key}'"
                ),
                "remove_hosts": _(
                    "Remove hosts from role '${primary_key}'"
                ),
                "remove_hostgroups": _(
                    "Remove host groups from role '${primary_key}'"
                ),
                "remove_services": _(
                    "Remove services from role '${primary_key}'"
                ),
                "remove_users": _(
                    "Remove users from role '${primary_key}'"
                ),
            },
            "aciselfservice": {
                "add": _("Add self service permission"),
                "remove": _("Remove self service permissions"),
            },
            "automember": {
                "add": _("Add rule"),
                "add_inc_condition": _(
                    "Add inclusive condition into '${primary_key}'"
                ),
                "add_exc_condition": _(
                    "Add exclusive condition into '${primary_key}'"
                ),
                "attribute": _("Attribute"),
                "default_host_group": _("Default host group"),
                "default_user_group": _("Default user group"),
                "exclusive": _("Exclusive"),
                "expression": _("Expression"),
                "hostgrouprule": _("Host group rule"),
                "hostgrouprules": _("Host group rules"),
                "inclusive": _("Inclusive"),
                "remove": _("Remove auto membership rules"),
                "remove_exc_conditions": _(
                    "Remove exclusive conditions from rule '${primary_key}'"
                ),
                "remove_inc_conditions": _(
                    "Remove inclusive conditions from rule '${primary_key}'"
                ),
                "usergrouprule": _("User group rule"),
                "usergrouprules": _("User group rules"),
            },
            "automountkey": {
                "add": _("Add automount key"),
                "remove": _("Remove automount keys"),
            },
            "automountlocation": {
                "add": _("Add automount location"),
                "identity": _("Automount Location Settings"),
                "remove": _("Remove automount locations"),
            },
            "automountmap": {
                "add": _("Add automount map"),
                "map_type": _("Map Type"),
                "direct": _("Direct"),
                "indirect": _("Indirect"),
                "remove": _("Remove automount maps"),
            },
            "ca": {
                "add": _("Add certificate authority"),
                "remove": _("Remove certificate authorities"),
            },
            "caacl": {
                "add": _("Add CA ACL"),
                "add_ca": _(
                    "Add Certificate Authorities into CA ACL "
                    "'${primary_key}'"
                ),
                "add_groups": _(
                    "Add user groups into CA ACL '${primary_key}'"
                ),
                "add_hostgroups": _(
                    "Add host groups into CA ACL '${primary_key}'"
                ),
                "add_hosts": _(
                    "Add hosts into CA ACL '${primary_key}'"
                ),
                "add_profiles": _(
                    "Add certificate profiles into CA ACL '${primary_key}'"
                ),
                "add_services": _(
                    "Add services into CA ACL '${primary_key}'"
                ),
                "add_users": _(
                    "Add users into CA ACL '${primary_key}'"
                ),
                "all": _("All"),
                "any_ca": _("Any CA"),
                "any_host": _("Any Host"),
                "any_service": _("Any Service"),
                "any_profile": _("Any Profile"),
                "anyone": _("Anyone"),
                "ipaenabledflag": _("Rule status"),
                "no_ca_msg": _("If no CAs are specified, requests to the default CA are allowed."),
                "profile": _("Profiles"),
                "remove": _("Remove CA ACLs"),
                "remove_ca": _(
                    "Remove Certificate Authorities from CA ACL "
                    "'${primary_key}'"
                ),
                "remove_groups": _(
                    "Remove user groups from CA ACL '${primary_key}'"
                ),
                "remove_hostgroups": _(
                    "Remove host groups from CA ACL '${primary_key}'"
                ),
                "remove_hosts": _(
                    "Remove hosts from CA ACL '${primary_key}'"
                ),
                "remove_profiles": _(
                    "Remove certificate profiles from CA ACL '${primary_key}'"
                ),
                "remove_services": _(
                    "Remove services from CA ACL '${primary_key}'"
                ),
                "remove_users": _(
                    "Remove users from CA ACL '${primary_key}'"
                ),
                "specified_cas": _("Specified CAs"),
                "specified_hosts": _("Specified Hosts and Groups"),
                "specified_profiles": _("Specified Profiles"),
                "specified_services": _("Specified Services and Groups"),
                "specified_users": _("Specified Users and Groups"),
                "who": _("Permitted to have certificates issued"),
            },
            "caprofile": {
                "remove": _("Remove certificate profiles"),
            },
            "cert": {
                "aa_compromise": _("AA Compromise"),
                "add_principal": _("Add principal"),
                "affiliation_changed": _("Affiliation Changed"),
                "ca": _("CA"),
                "ca_compromise": _("CA Compromise"),
                "certificate": _("Certificate"),
                "certificates": _("Certificates"),
                "certificate_hold": _("Certificate Hold"),
                "cessation_of_operation": _("Cessation of Operation"),
                "common_name": _("Common Name"),
                "download": _("Download"),
                "delete_cert_end": _("the certificate with serial number "),
                "expires_on": _("Expires On"),
                "find_issuedon_from": _("Issued on from"),
                "find_issuedon_to": _("Issued on to"),
                "find_max_serial_number": _("Maximum serial number"),
                "find_min_serial_number": _("Minimum serial number"),
                "find_revocation_reason": _("Revocation reason"),
                "find_revokedon_from": _("Revoked on from"),
                "find_revokedon_to": _("Revoked on to"),
                "find_subject": _("Subject"),
                "find_validnotafter_from": _("Valid not after from"),
                "find_validnotafter_to": _("Valid not after to"),
                "find_validnotbefore_from": _("Valid not before from"),
                "find_validnotbefore_to": _("Valid not before to"),
                "fingerprints": _("Fingerprints"),
                "get_certificate": _("Get Certificate"),
                "hold_removed": _("Certificate Hold Removed"),
                "issue_for_host": _(
                    "Issue new certificate for host '${primary_key}'"
                ),
                "issue_for_service": _(
                    "Issue new certificate for service '${primary_key}'"
                ),
                "issue_for_user": _(
                    "Issue new certificate for user '${primary_key}'"
                ),
                "issue_certificate_generic": _("Issue new certificate"),
                "issued_by": _("Issued By"),
                "issued_on": _("Issued On"),
                "issued_to": _("Issued To"),
                "key_compromise": _("Key Compromise"),
                "missing": _("No Valid Certificate"),
                "new_certificate": _("New Certificate"),
                "new_cert_format": _("Certificate in base64 or PEM format"),
                "note": _("Note"),
                "organization": _("Organization"),
                "organizational_unit": _("Organizational Unit"),
                "present": _("${count} certificate(s) present"),
                "privilege_withdrawn": _("Privilege Withdrawn"),
                "reason": _("Reason for Revocation"),
                "remove_hold": _("Remove hold"),
                "remove_certificate_hold_simple": _("Remove certificate hold"),
                "remove_certificate_hold_confirmation": _("Do you want to remove the certificate hold?"),
                "remove_from_crl": _("Remove from CRL"),
                "request_message": _("<ol> <li>Create a certificate database or use an existing one. To create a new database:<br/> <code># certutil -N -d &lt;database path&gt;</code> </li> <li>Create a CSR with subject <em>CN=&lt;${cn_name}&gt;,O=&lt;realm&gt;</em>, for example:<br/> <code># certutil -R -d &lt;database path&gt; -a -g &lt;key size&gt; -s 'CN=${cn},O=${realm}'${san}</code> </li> <li> Copy and paste the CSR (from <em>-----BEGIN NEW CERTIFICATE REQUEST-----</em> to <em>-----END NEW CERTIFICATE REQUEST-----</em>) into the text area below: </li> </ol>"),
                "request_message_san": _(" -8 '${cn}'"),
                "requested": _("Certificate requested"),
                "revocation_reason": _("Revocation reason"),
                "revoke_certificate_simple": _("Revoke certificate"),
                "revoke_confirmation": _("Do you want to revoke this certificate? Select a reason from the pull-down list."),
                "revoked": _("Certificate Revoked"),
                "revoked_status": _("REVOKED"),
                "serial_number": _("Serial Number"),
                "serial_number_hex": _("Serial Number (hex)"),
                "sha1_fingerprint": _("SHA1 Fingerprint"),
                "sha256_fingerprint": _("SHA256 Fingerprint"),
                "status": _("Status"),
                "superseded": _("Superseded"),
                "unspecified": _("Unspecified"),
                "valid": _("Valid Certificate Present"),
                "valid_from": _("Valid from"),
                "valid_to": _("Valid to"),
                "validity": _("Validity"),
                "view_certificate": _("Certificate for ${entity} ${primary_key}"),
                "view_certificate_btn": _("View Certificate"),
            },
            "certmap_match": {
                "cert_data": _("Certificate Data"),
                "cert_for_match": _("Certificate For Match"),
                "facet_label": _("Certificate Mapping Match"),
                "domain": _("Domain"),
                "matched_users": _("Matched Users"),
                "userlogin": _("User Login"),
            },
            "certmap": {
                "add": _("Add certificate identity mapping rule"),
                "adder_title": _("Add certificate mapping data"),
                "data_label": _("Certificate mapping data"),
                "certificate": _("Certificate"),
                "conf_str": _("Configuration string"),
                "deleter_content": _("Do you want to remove certificate mapping data ${data}?"),
                "deleter_title": _("Remove certificate mapping data"),
                "issuer": _("Issuer"),
                "issuer_subject": _("Issuer and subject"),
                "remove": _("Remove certificate identity mapping rules"),
                "subject": _("Subject"),
                "version": _("Version"),
            },
            "config": {
                "group": _("Group Options"),
                "search": _("Search Options"),
                "selinux": _("SELinux Options"),
                "service": _("Service Options"),
                "user": _("User Options"),
            },
            "delegation": {
            },
            "dnsconfig": {
                "forward_first": _("Forward first"),
                "forward_none": _("Forwarding disabled"),
                "forward_only": _("Forward only"),
                "options": _("Options"),
                "update_dns": _("Update System DNS Records"),
                "update_dns_dialog_msg": _("Do you want to update system DNS records?"),
                "updated_dns": _("System DNS records updated"),
            },
            "dnsforwardzone": {
                "add": _("Add DNS forward zone"),
                "remove": _("Remove DNS forward zones"),
            },
            "dnsrecord": {
                "add": _("Add DNS resource record"),
                "data": _("Data"),
                "deleted_no_data": _("DNS record was deleted because it contained no data."),
                "other": _("Other Record Types"),
                "ptr_redir_address_err": _("Address not valid, can't redirect"),
                "ptr_redir_create": _("Create dns record"),
                "ptr_redir_creating": _("Creating record."),
                "ptr_redir_creating_err": _("Record creation failed."),
                "ptr_redir_record": _("Checking if record exists."),
                "ptr_redir_record_err": _("Record not found."),
                "ptr_redir_title": _("Redirection to PTR record"),
                "ptr_redir_zone": _("Zone found: ${zone}"),
                "ptr_redir_zone_err": _("Target reverse zone not found."),
                "ptr_redir_zones": _("Fetching DNS zones."),
                "ptr_redir_zones_err": _("An error occurred while fetching dns zones."),
                "redirection_dnszone": _("You will be redirected to DNS Zone."),
                "remove": _("Remove DNS resource records"),
                "standard": _("Standard Record Types"),
                "title": _("Records for DNS Zone"),
                "type": _("Record Type"),
            },
            "dnszone": {
                "add": _("Add DNS zone"),
                "add_permission": _("Add permission"),
                "add_permission_confirm":_("Are you sure you want to add permission for DNS Zone ${object}?"),
                "identity": _("DNS Zone Settings"),
                "remove": _("Remove DNS zones"),
                "remove_permission": _("Remove Permission"),
                "remove_permission_confirm": _("Are you sure you want to remove permission for DNS Zone ${object}?"),
                "skip_dns_check": _("Skip DNS check"),
                "skip_overlap_check": _("Skip overlap check"),
                "soamname_change_message": _("Do you want to check if new authoritative nameserver address is in DNS"),
                "soamname_change_title": _("Authoritative nameserver change"),
            },
            "domainlevel": {
                "label": _("Domain Level"),
                "label_singular": _("Domain Level"),
                "ipadomainlevel": _("Level"),
                "set": _("Set Domain Level"),
            },
            "group": {
                "add": _("Add user group"),
                "add_groups": _(
                    "Add user groups into user group '${primary_key}'"
                ),
                "add_into_groups": _(
                    "Add user group '${primary_key}' into user groups"
                ),
                "add_into_hbac": _(
                    "Add user group '${primary_key}' into HBAC rules"
                ),
                "add_into_netgroups": _(
                    "Add user group '${primary_key}' into netgroups"
                ),
                "add_into_roles": _(
                    "Add user group '${primary_key}' into roles"
                ),
                "add_into_sudo": _(
                    "Add user group '${primary_key}' into sudo rules"
                ),
                "add_services": _(
                    "Add services into user group '${primary_key}'"
                ),
                "add_users": _(
                    "Add users into user group '${primary_key}'"
                ),
                "details": _("Group Settings"),
                "external": _("External"),
                "groups": _("Groups"),
                "group_categories": _("Group categories"),
                "make_external": _("Change to external group"),
                "make_posix": _("Change to POSIX group"),
                "nonposix": _("Non-POSIX"),
                "posix": _("POSIX"),
                "remove": _("Remove user groups"),
                "remove_from_groups": _(
                    "Remove user group '${primary_key}' from user groups"
                ),
                "remove_from_netgroups": _(
                    "Remove user group '${primary_key}' from netgroups"
                ),
                "remove_from_roles": _(
                    "Remove user group '${primary_key}' from roles"
                ),
                "remove_from_hbac": _(
                    "Remove user group '${primary_key}' from HBAC rules"
                ),
                "remove_from_sudo": _(
                    "Remove user group '${primary_key}' from sudo rules"
                ),
                "remove_groups": _(
                    "Remove user groups from user group '${primary_key}'"
                ),
                "remove_services": _(
                    "Remove services from user group '${primary_key}'"
                ),
                "remove_users": _(
                    "Remove users from user group '${primary_key}'"
                ),
                "type": _("Group Type"),
                "user_groups": _("User Groups"),
            },
            "hbacrule": {
                "add": _("Add HBAC rule"),
                "add_groups": _(
                    "Add user groups into HBAC rule '${primary_key}'"
                ),
                "add_hostgroups": _(
                    "Add host groups into HBAC rule '${primary_key}'"
                ),
                "add_hosts": _(
                    "Add hosts into HBAC rule '${primary_key}'"
                ),
                "add_servicegroups": _(
                    "Add HBAC service groups into HBAC rule "
                    "'${primary_key}'"
                ),
                "add_services": _(
                    "Add HBAC services into HBAC rule '${primary_key}'"
                ),
                "add_users": _(
                    "Add users into HBAC rule '${primary_key}'"
                ),
                "any_host": _("Any Host"),
                "any_service": _("Any Service"),
                "anyone": _("Anyone"),
                "host": _("Accessing"),
                "ipaenabledflag": _("Rule status"),
                "remove": _("Remove HBAC rules"),
                "remove_groups": _(
                    "Remove user groups from HBAC rule '${primary_key}'"
                ),
                "remove_hostgroups": _(
                    "Remove host groups from HBAC rule '${primary_key}'"
                ),
                "remove_hosts": _(
                    "Remove hosts from HBAC rule '${primary_key}'"
                ),
                "remove_servicegroups": _(
                    "Remove HBAC service groups from HBAC rule "
                    "'${primary_key}'"
                ),
                "remove_services": _(
                    "Remove HBAC services from HBAC rule '${primary_key}'"
                ),
                "remove_users": _(
                    "Remove users from HBAC rule '${primary_key}'"
                ),
                "service": _("Via Service"),
                "specified_hosts": _("Specified Hosts and Groups"),
                "specified_services": _("Specified Services and Groups"),
                "specified_users": _("Specified Users and Groups"),
                "user": _("Who"),
            },
            "hbacsvc": {
                "add": _("Add HBAC service"),
                "add_hbacsvcgroups": _(
                    "Add HBAC service '${primary_key}' into HBAC service "
                    "groups"
                ),
                "remove": _("Remove HBAC services"),
                "remove_from_hbacsvcgroups": _(
                    "Remove HBAC service '${primary_key}' from HBAC service "
                    "groups"
                ),
            },
            "hbacsvcgroup": {
                "add": _("Add HBAC service group"),
                "add_hbacsvcs": _(
                    "Add HBAC services into HBAC service group "
                    "'${primary_key}'"
                ),
                "remove": _("Remove HBAC service groups"),
                "remove_hbacsvcs": _(
                    "Remove HBAC services from HBAC service group "
                    "'${primary_key}'"
                ),
                "services": _("Services"),
            },
            "hbactest": {
                "access_denied": _("Access Denied"),
                "access_granted": _("Access Granted"),
                "include_disabled": _("Include Disabled"),
                "include_enabled": _("Include Enabled"),
                "label": _("HBAC Test"),
                "matched": _("Matched"),
                "missing_values": _("Missing values: "),
                "new_test": _("New Test"),
                "rules": _("Rules"),
                "run_test": _("Run Test"),
                "specify_external": _("Specify external ${entity}"),
                "unmatched": _("Unmatched"),
            },
            "host": {
                "add": _("Add host"),
                "add_hosts_managing": _(
                    "Add hosts managing host '${primary_key}'"
                ),
                "add_into_groups": _(
                    "Add host '${primary_key}' into host groups"
                ),
                "add_into_hbac": _(
                    "Add host '${primary_key}' into HBAC rules"
                ),
                "add_into_netgroups": _(
                    "Add host '${primary_key}' into netgroups"
                ),
                "add_into_roles": _(
                    "Add host '${primary_key}' into roles"
                ),
                "add_into_sudo": _(
                    "Add host '${primary_key}' into sudo rules"
                ),
                "certificate": _("Host Certificate"),
                "cn": _("Host Name"),
                "delete_key_unprovision": _("Delete Key, Unprovision"),
                "details": _("Host Settings"),
                "enrolled": _("Enrolled"),
                "enrollment": _("Enrollment"),
                "fqdn": _("Fully Qualified Host Name"),
                "generate_otp": _("Generate OTP"),
                "generated_otp": _("Generated OTP"),
                "keytab": _("Kerberos Key"),
                "keytab_missing": _("Kerberos Key Not Present"),
                "keytab_present": _("Kerberos Key Present, Host Provisioned"),
                "password": _("One-Time-Password"),
                "password_missing": _("One-Time-Password Not Present"),
                "password_present": _("One-Time-Password Present"),
                "password_reset_button": _("Reset OTP"),
                "password_reset_title": _("Reset One-Time-Password"),
                "password_set_button": _("Set OTP"),
                "password_set_success": _("OTP set"),
                "password_set_title": _("Set One-Time-Password"),
                "remove": _("Remove hosts"),
                "remove_hosts_managing": _(
                    "Remove hosts managing host '${primary_key}'"
                ),
                "remove_from_groups": _(
                    "Remove host '${primary_key}' from host groups"
                ),
                "remove_from_netgroups": _(
                    "Remove host '${primary_key}' from netgroups"
                ),
                "remove_from_roles": _(
                    "Remove host '${primary_key}' from roles"
                ),
                "remove_from_hbac": _(
                    "Remove host '${primary_key}' from HBAC rules"
                ),
                "remove_from_sudo": _(
                    "Remove host '${primary_key}' from sudo rules"
                ),
                "status": _("Status"),
                "unprovision": _("Unprovision"),
                "unprovision_confirmation": _("Are you sure you want to unprovision this host?"),
                "unprovision_title": _("Unprovisioning host"),
                "unprovisioned": _("Host unprovisioned"),
            },
            "hostgroup": {
                "add": _("Add host group"),
                "add_hosts": _(
                    "Add hosts into host group '${primary_key}'"
                ),
                "add_hostgroups": _(
                    "Add host groups into host group '${primary_key}'"
                ),
                "add_into_hostgroups": _(
                    "Add host group '${primary_key}' into host groups"
                ),
                "add_into_hbac": _(
                    "Add host group '${primary_key}' into HBAC rules"
                ),
                "add_into_netgroups": _(
                    "Add host group '${primary_key}' into netgroups"
                ),
                "add_into_sudo": _(
                    "Add host group '${primary_key}' into sudo rules"
                ),
                "host_group": _("Host Groups"),
                "identity": _("Host Group Settings"),
                "remove": _("Remove host groups"),
                "remove_from_hostgroups": _(
                    "Remove host group '${primary_key}' from host groups"
                ),
                "remove_from_netgroups": _(
                    "Remove host group '${primary_key}' from netgroups"
                ),
                "remove_from_hbac": _(
                    "Remove host group '${primary_key}' from HBAC rules"
                ),
                "remove_from_sudo": _(
                    "Remove host group '${primary_key}' from sudo rules"
                ),
                "remove_hosts": _(
                    "Remove hosts from host group '${primary_key}'"
                ),
                "remove_hostgroups": _(
                    "Remove host groups from host group '${primary_key}'"
                ),
            },
            "idoverrideuser": {
                "anchor_label": _("User to override"),
                "anchor_tooltip": _("Enter trusted or IPA user login. Note: search doesn't list users from trusted domains."),
                "anchor_tooltip_ad": _("Enter trusted user login."),
                "profile": _("Profile"),
            },
            "idoverridegroup": {
                "anchor_label": _("Group to override"),
                "anchor_tooltip": _("Enter trusted or IPA group name. Note: search doesn't list groups from trusted domains."),
                "anchor_tooltip_ad": _("Enter trusted group name."),
            },
            "idview": {
                "add": _("Add ID view"),
                "add_group": _("Add group ID override"),
                "add_user": _("Add user ID override"),
                "appliesto_tab": _("${primary_key} applies to:"),
                "appliedtohosts": _("Applied to hosts"),
                "appliedtohosts_title": _("Applied to hosts"),
                "apply_hostgroups": _("Apply to host groups"),
                "apply_hostgroups_title": _(
                    "Apply ID view '${primary_key}' on hosts of host groups"
                ),
                "apply_hosts": _("Apply to hosts"),
                "apply_hosts_title": _(
                    "Apply ID view '${primary_key}' on hosts"
                ),
                "ipaassignedidview": _("Assigned ID View"),
                "overrides_tab": _("${primary_key} overrides:"),
                "remove": _("Remove ID views"),
                "remove_users": _("Remove user ID overrides"),
                "remove_groups": _("Remove group ID overrides"),
                "unapply_hostgroups": _("Un-apply from host groups"),
                "unapply_hostgroups_all_title": _("Un-apply ID Views from hosts of hostgroups"),
                "unapply_hosts": _("Un-apply"),
                "unapply_hosts_all": _("Un-apply from hosts"),
                "unapply_hosts_all_title": _("Un-apply ID Views from hosts"),
                "unapply_hosts_confirm": _("Are you sure you want to un-apply ID view from selected entries?"),
                "unapply_hosts_title": _(
                    "Un-apply ID view '${primary_key}' from hosts"
                ),
            },
            "krbtpolicy": {
                "identity": _("Kerberos Ticket Policy"),
            },
            "netgroup": {
                "add": _("Add netgroup"),
                "add_into_netgroups": _(
                    "Add netgroup '${primary_key}' into netgroups"
                ),
                "add_netgroups": _(
                    "Add netgroups into netgroup '${primary_key}'"
                ),
                "add_groups": _(
                    "Add user groups into netgroup '${primary_key}'"
                ),
                "add_hosts": _(
                    "Add hosts into netgroup '${primary_key}'"
                ),
                "add_hostgroups": _(
                    "Add host groups into netgroup '${primary_key}'"
                ),
                "add_users": _(
                    "Add users into netgroup '${primary_key}'"
                ),
                "any_host": _("Any Host"),
                "anyone": _("Anyone"),
                "external": _("External"),
                "host": _("Host"),
                "hostgroups": _("Host Groups"),
                "hosts": _("Hosts"),
                "identity": _("Netgroup Settings"),
                "netgroups": _("Netgroups"),
                "remove": _("Remove netgroups"),
                "remove_from_netgroups": _(
                    "Remove netgroup '${primary_key}' from netgroups"
                ),
                "remove_groups": _(
                    "Remove user groups from netgroup '${primary_key}'"
                ),
                "remove_hosts": _(
                    "Remove hosts from netgroup '${primary_key}'"
                ),
                "remove_hostgroups": _(
                    "Remove host groups from netgroup '${primary_key}'"
                ),
                "remove_netgroups": _(
                    "Remove netgroups from netgroup '${primary_key}'"
                ),
                "remove_users": _(
                    "Remove users from netgroup '${primary_key}'"
                ),
                "specified_hosts": _("Specified Hosts and Groups"),
                "specified_users": _("Specified Users and Groups"),
                "user": _("User"),
                "usergroups": _("User Groups"),
                "users": _("Users"),
            },
            "otptoken": {
                "add": _("Add OTP token"),
                "add_users_managing": _(
                    "Add users managing OTP token '${primary_key}'"
                ),
                "app_link": _("You can use <a href=\"${link}\" target=\"_blank\">FreeOTP<a/> as a software OTP token application."),
                "config_title": _("Configure your token"),
                "config_instructions": _("Configure your token by scanning the QR code below. Click on the QR code if you see this on the device you want to configure."),
                "details": _("OTP Token Settings"),
                "disable": _("Disable token"),
                "enable": _("Enable token"),
                "remove": _("Remove OTP tokens"),
                "remove_users_managing": _(
                    "Remove users managing OTP token '${primary_key}'"
                ),
                "show_qr": _("Show QR code"),
                "show_uri": _("Show configuration uri"),
                "type_hotp": _("Counter-based (HOTP)"),
                "type_totp": _("Time-based (TOTP)"),
            },
            "permission": {
                "add_custom_attr": _("Add Custom Attribute"),
                "attribute": _("Attribute"),
                "filter": _("Filter"),
                "identity": _("Permission settings"),
                "managed": _("Attribute breakdown"),
                "target": _("Target"),
            },
            "privilege": {
                "identity": _("Privilege Settings"),
            },
            "publickey": {
                "set_dialog_help": _("Public key:"),
                "set_dialog_title": _("Set public key"),
                "show_set_key": _("Show/Set key"),
                "status_mod_ns": _("Modified: key not set"),
                "status_mod_s": _("Modified"),
                "status_new_ns": _("New: key not set"),
                "status_new_s": _("New: key set"),
            },
            "pwpolicy": {
                "add": _("Add password policy"),
                "identity": _("Password Policy"),
                "remove": _("Remove password policies"),
            },
            "idrange": {
                "add": _("Add ID range"),
                "details": _("Range Settings"),
                "ipabaseid": _("Base ID"),
                "ipabaserid": _("Primary RID base"),
                "ipaidrangesize": _("Range size"),
                "ipanttrusteddomainsid": _("Domain SID"),
                "ipasecondarybaserid": _("Secondary RID base"),
                "remove": _("Remove ID ranges"),
                "type": _("Range type"),
                "type_ad": _("Active Directory domain"),
                "type_ad_posix": _("Active Directory domain with POSIX attributes"),
                "type_detect": _("Detect"),
                "type_local": _("Local domain"),
                "type_ipa": _("IPA trust"),
                "type_winsync": _("Active Directory winsync"),
            },
            "radiusproxy": {
                "add": _("Add RADIUS server"),
                "details": _("RADIUS Proxy Server Settings"),
                "remove": _("Remove RADIUS servers"),
            },
            "realmdomains": {
                "identity": _("Realm Domains"),
                "check_dns": _("Check DNS"),
                "check_dns_confirmation": _("Do you also want to perform DNS check?"),
                "force_update": _("Force Update"),
            },
            "selfservice": {
            },
            "selinuxusermap": {
                "add": _("Add SELinux user map"),
                "add_groups": _(
                    "Add user groups into SELinux user map '${primary_key}'"
                ),
                "add_hostgroups": _(
                    "Add host groups into SELinux user map '${primary_key}'"
                ),
                "add_hosts": _(
                    "Add hosts into SELinux user map '${primary_key}'"
                ),
                "add_users": _(
                    "Add users into SELinux user map '${primary_key}'"
                ),
                "any_host": _("Any Host"),
                "anyone": _("Anyone"),
                "host": _("Host"),
                "remove": _("Remove selinux user maps"),
                "remove_groups": _(
                    "Remove user groups from SELinux user map '${primary_key}'"
                ),
                "remove_hostgroups": _(
                    "Remove host groups from SELinux user map '${primary_key}'"
                ),
                "remove_hosts": _(
                    "Remove hosts from SELinux user map '${primary_key}'"
                ),
                "remove_users": _(
                    "Remove users from SELinux user map '${primary_key}'"
                ),
                "specified_hosts": _("Specified Hosts and Groups"),
                "specified_users": _("Specified Users and Groups"),
                "user": _("User"),
            },
            "server_role": {
                "label": _("Server Roles"),
                "label_singular": _("Server Role"),
            },
            "servers": {
                "svc_warning_title": _("Warning: Consider service replication"),
                "svc_warning_message": _("It is strongly recommended to keep the following services installed on more than one server:"),
                "remove_server": _("Delete Server"),
                "remove_server_msg": _("Deleting a server removes it permanently from the topology. Note that this is a non-reversible action.")
            },
            "service": {
                "add": _("Add service"),
                "add_hosts_managing": _(
                    "Add hosts managing service '${primary_key}'"
                ),
                "add_into_roles": _(
                    "Add service '${primary_key}' into roles"
                ),
                "certificate": _("Service Certificate"),
                "delete_key_unprovision": _("Delete Key, Unprovision"),
                "details": _("Service Settings"),
                "host": _("Host Name"),
                "missing": _("Kerberos Key Not Present"),
                "provisioning": _("Provisioning"),
                "remove": _("Remove services"),
                "remove_from_roles": _(
                    "Remove service '${primary_key}' from roles"
                ),
                "remove_hosts_managing": _(
                    "Remove hosts managing service '${primary_key}'"
                ),
                "service": _("Service"),
                "status": _("Status"),
                "unprovision": _("Unprovision"),
                "unprovision_confirmation": _("Are you sure you want to unprovision this service?"),
                "unprovision_title": _("Unprovisioning service"),
                "unprovisioned": _("Service unprovisioned"),
                "valid": _("Kerberos Key Present, Service Provisioned"),
            },
            "sshkeystore": {
                "keys": _("SSH public keys"),
                "set_dialog_help": _("SSH public key:"),
                "set_dialog_title": _("Set SSH key"),
                "show_set_key": _("Show/Set key"),
                "status_mod_ns": _("Modified: key not set"),
                "status_mod_s": _("Modified"),
                "status_new_ns": _("New: key not set"),
                "status_new_s": _("New: key set"),
            },
            "stageuser": {
                "activate_confirm": _("Are you sure you want to activate selected users?"),
                "activate_one_confirm": _("Are you sure you want to activate ${object}?"),
                "activate_success": _("${count} user(s) activated"),
                "add": _("Add stage user"),
                "label": _("Stage users"),
                "preserved_label": _("Preserved users"),
                "preserved_remove": _("Remove preserved users"),
                "remove": _("Remove stage users"),
                "stage_confirm": _("Are you sure you want to stage selected users?"),
                "stage_success": _("${count} users(s) staged"),
                "stage_one_confirm": _("Are you sure you want to stage ${object}?"),
                "undel_confirm": _("Are you sure you want to restore selected users?"),
                "undel_one_confirm": _("Are you sure you want to restore ${object}?"),
                "undel_success": _("${count} user(s) restored"),
                "user_categories": _("User categories"),
            },
            "sudocmd": {
                "add": _("Add sudo command"),
                "add_into_sudocmdgroups": _(
                    "Add sudo command '${primary_key}' into sudo command "
                    "groups"
                ),
                "groups": _("Groups"),
                "remove": _("Remove sudo commands"),
                "remove_from_sudocmdgroups": _(
                    "Remove sudo command '${primary_key}' from sudo command "
                    "groups"
                ),
            },
            "sudocmdgroup": {
                "add": _("Add sudo command group"),
                "add_sudocmds": _(
                    "Add sudo commands into sudo command group "
                    "'${primary_key}'"
                ),
                "commands": _("Commands"),
                "remove": _("Remove sudo command groups"),
                "remove_sudocmds": _(
                    "Remove sudo commands from sudo command group "
                    "'${primary_key}'"
                ),
            },
            "sudorule": {
                "add": _("Add sudo rule"),
                "add_option": _("Add sudo option"),
                "add_allow_cmds": _(
                    "Add allow sudo commands into sudo rule "
                    "'${primary_key}'"
                ),
                "add_allow_cmdgroups": _(
                    "Add allow sudo command groups into sudo rule "
                    "'${primary_key}'"
                ),
                "add_deny_cmds": _(
                    "Add deny sudo commands into sudo rule "
                    "'${primary_key}'"
                ),
                "add_deny_cmdgroups": _(
                    "Add deny sudo command groups into sudo rule "
                    "'${primary_key}'"
                ),
                "add_groups": _(
                    "Add user groups into sudo rule '${primary_key}'"
                ),
                "add_hostgroups": _(
                    "Add host groups into sudo rule '${primary_key}'"
                ),
                "add_hosts": _(
                    "Add hosts into sudo rule '${primary_key}'"
                ),
                "add_runas_users": _(
                    "Add RunAs users into sudo rule '${primary_key}'"
                ),
                "add_runas_usergroups": _(
                    "Add RunAs user groups into sudo rule '${primary_key}'"
                ),
                "add_runas_groups": _(
                    "Add RunAs groups into sudo rule '${primary_key}'"
                ),
                "add_users": _(
                    "Add users into sudo rule '${primary_key}'"
                ),
                "allow": _("Allow"),
                "any_command": _("Any Command"),
                "any_group": _("Any Group"),
                "any_host": _("Any Host"),
                "anyone": _("Anyone"),
                "command": _("Run Commands"),
                "deny": _("Deny"),
                "external": _("External"),
                "host": _("Access this host"),
                "ipaenabledflag": _("Rule status"),
                "option_added": _("Option added"),
                "option_removed": _("${count} option(s) removed"),
                "options": _("Options"),
                "remove": _("Remove sudo rules"),
                "remove_allow_cmds": _(
                    "Remove allow sudo commands from sudo rule "
                    "'${primary_key}'"
                ),
                "remove_allow_cmdgroups": _(
                    "Remove allow sudo command groups from sudo rule "
                    "'${primary_key}'"
                ),
                "remove_deny_cmds": _(
                    "Remove deny sudo commands from sudo rule "
                    "'${primary_key}'"
                ),
                "remove_deny_cmdgroups": _(
                    "Remove deny sudo command groups from sudo rule "
                    "'${primary_key}'"
                ),
                "remove_groups": _(
                    "Remove user groups from sudo rule '${primary_key}'"
                ),
                "remove_hostgroups": _(
                    "Remove host groups from sudo rule '${primary_key}'"
                ),
                "remove_hosts": _(
                    "Remove hosts from sudo rule '${primary_key}'"
                ),
                "remove_runas_users": _(
                    "Remove RunAs users from sudo rule '${primary_key}'"
                ),
                "remove_runas_usergroups": _(
                    "Remove RunAs user groups from sudo rule '${primary_key}'"
                ),
                "remove_runas_groups": _(
                    "Remove RunAs groups from sudo rule '${primary_key}'"
                ),
                "remove_users": _(
                    "Remove users from sudo rule '${primary_key}'"
                ),
                "runas": _("As Whom"),
                "specified_commands": _("Specified Commands and Groups"),
                "specified_groups": _("Specified Groups"),
                "specified_hosts": _("Specified Hosts and Groups"),
                "specified_users": _("Specified Users and Groups"),
                "user": _("Who"),
            },
            "sudooptions": {
                "remove": _("Remove sudo options"),
            },
            "topology": {
                "autogenerated": _("Autogenerated"),
                "segment_details": _("Segment details"),
                "replication_config": _("Replication configuration"),
                "insufficient_domain_level" : _("Managed topology requires minimal domain level ${domainlevel}"),
            },
            "topologylocation": {
                "add": _("Add IPA location"),
                "add_server": _(
                    "Add IPA server into IPA location '${primary_key}'"
                ),
                "remove": _("Remove IPA locations"),
                "remove_servers": _(
                    "Remove IPA servers from IPA location '${primary_key}'"
                ),
            },
            "topologysegment": {
                "add": _("Add topology segment"),
                "remove": _("Remove topology segments"),
            },
            "trust": {
                "account": _("Account"),
                "add": _("Add trust"),
                "admin_account": _("Administrative account"),
                "blacklists": _("SID blacklists"),
                "details": _("Trust Settings"),
                "domain": _("Domain"),
                "establish_using": _("Establish using"),
                "fetch_domains": _("Fetch domains"),
                "ipantflatname": _("Domain NetBIOS name"),
                "ipanttrusteddomainsid": _("Domain Security Identifier"),
                "preshared_password": _("Pre-shared password"),
                "remove": _("Remove trusts"),
                "remove_domains": _("Remove domains"),
                "trustdirection": _("Trust direction"),
                "truststatus": _("Trust status"),
                "trusttype": _("Trust type"),
                "ipantadditionalsuffixes": _("Alternative UPN suffixes"),
            },
            'smb_attributes': {
                "title": _(
                    "User attributes for SMB services"
                ),
                "ipantlogonscript_tooltip": _(
                    "Path to a script executed on a Windows system at logon"
                ),
                "ipantprofilepath_tooltip": _(
                    "Path to a user profile, in UNC format \\\\server\\share\\"
                ),
                "ipanthomedirectory_tooltip": _(
                    "Path to a user home directory, in UNC format"
                ),
                "ipanthomedirectorydrive_tooltip": _(
                    "Drive to mount a home directory"
                ),
            },
            "trustconfig": {
                "options": _("Options"),
            },
            "user": {
                "account": _("Account Settings"),
                "account_status": _("Account Status"),
                "activeuser_label": _("Active users"),
                "add": _("Add user"),
                "add_into_groups": _(
                    "Add user '${primary_key}' into user groups"
                ),
                "add_into_hbac": _(
                    "Add user '${primary_key}' into HBAC rules"
                ),
                "add_into_netgroups": _(
                    "Add user '${primary_key}' into netgroups"
                ),
                "add_into_roles": _(
                    "Add user '${primary_key}' into roles"
                ),
                "add_into_sudo": _(
                    "Add user '${primary_key}' into sudo rules"
                ),
                "contact": _("Contact Settings"),
                "delete_mode": _("Delete mode"),
                "employee": _("Employee Information"),
                "error_changing_status": _("Error changing account status"),
                "krbpasswordexpiration": _("Password expiration"),
                "mailing": _("Mailing Address"),
                "misc": _("Misc. Information"),
                "mode_delete": _("delete"),
                "mode_preserve": _("preserve"),
                "noprivate": _("No private group"),
                "remove": _("Remove users"),
                "remove_from_groups": _(
                    "Remove user '${primary_key}' from user groups"
                ),
                "remove_from_netgroups": _(
                    "Remove user '${primary_key}' from netgroups"
                ),
                "remove_from_roles": _(
                    "Remove user '${primary_key}' from roles"
                ),
                "remove_from_hbac": _(
                    "Remove user '${primary_key}' from HBAC rules"
                ),
                "remove_from_sudo": _(
                    "Remove user '${primary_key}' from sudo rules"
                ),
                "status_confirmation": _("Are you sure you want to ${action} the user?<br/>The change will take effect immediately."),
                "status_link": _("Click to ${action}"),
                "unlock": _("Unlock"),
                "unlock_confirm": _("Are you sure you want to unlock user ${object}?"),
            },
            "vault": {
                "add": _("Add vault"),
                "add_member_groups": _(
                    "Add user groups into members of vault '${primary_key}'"
                ),
                "add_member_services": _(
                    "Add services into members of vault '${primary_key}'"
                ),
                "add_member_users": _(
                    "Add users into members of vault '${primary_key}'"
                ),
                "add_owner_groups": _(
                    "Add user groups into owners of vault '${primary_key}'"
                ),
                "add_owner_services": _(
                    "Add services into owners of vault '${primary_key}'"
                ),
                "add_owner_users": _(
                    "Add users into owners of vault '${primary_key}'"
                ),
                "add_warn_arch_ret": _(
                    "Secrets can be added/retrieved to vault only by using "
                    "vault-archive and vault-retrieve from CLI."
                    ),
                "add_warn_standard": _(
                    "Content of 'standard' vaults can be seen by users with "
                    "higher privileges (admins)."
                    ),
                "asymmetric_type": _("Asymmetric"),
                "config_title": _("Vaults Config"),
                "group": _("Group"),
                "members": _("Members"),
                "my_vaults_title": _("My User Vaults"),
                "owners": _("Owners"),
                "remove": _("Remove vaults"),
                "remove_member_groups": _(
                    "Remove user groups from members of vault '${primary_key}'"
                ),
                "remove_member_services": _(
                    "Remove services from members of vault '${primary_key}'"
                ),
                "remove_member_users": _(
                    "Remove users from members of vault '${primary_key}'"
                ),
                "remove_owner_groups": _(
                    "Remove user groups from owners of vault '${primary_key}'"
                ),
                "remove_owner_services": _(
                    "Remove services from owners of vault '${primary_key}'"
                ),
                "remove_owner_users": _(
                    "Remove users from owners of vault '${primary_key}'"
                ),
                "service": _("Service"),
                "service_vaults_title": _("Service Vaults"),
                "shared": _("Shared"),
                "shared_vaults_title": _("Shared Vaults"),
                "standard_type": _("Standard"),
                "symmetric_type": _("Symmetric"),
                "type": _("Vault Type"),
                "type_tooltip": _(
                    "Only standard vaults can be created in WebUI, use CLI "
                    "for other types of vaults."
                    ),
                "user": _("User"),
                "user_vaults_title": _("User Vaults"),
            },
        },
        "password": {
            "current_password": _("Current Password"),
            "current_password_required": _("Current password is required"),
            "expires_in": _("Your password expires in ${days} days."),
            "first_otp": _("First OTP"),
            "invalid_password": _(
                "The password or username you entered is incorrect"),
            "new_password": _("New Password"),
            "new_password_required": _("New password is required"),
            "otp": _("OTP"),
            "otp_info": _("<i class=\"fa fa-info-circle\"></i> <strong>One-Time-Password(OTP):</strong> Generate new OTP code for each OTP field."),
            "otp_long": _("One-Time-Password"),
            "otp_sync_fail": _("Token synchronization failed"),
            "otp_sync_invalid": _("The username, password or token codes are not correct"),
            "otp_sync_success":_("Token was synchronized"),
            "password": _("Password"),
            "password_and_otp": _("Password or Password+One-Time-Password"),
            "password_change_complete": _("Password change complete"),
            "password_expired": _(
                "Your password has expired. Please enter a new password."),
            "password_must_match": _("Passwords must match"),
            "reset_failure": _("Password reset was not successful."),
            "reset_password": _("Reset Password"),
            "reset_password_sentence": _("Reset your password."),
            "second_otp": _("Second OTP"),
            "token_id": _("Token ID"),
            "verify_password": _("Verify Password"),
        },
        "profile-menu": {
            "about": _("About"),
            "configuration": _("Customization"),
            "logout": _("Log out"),
            "password_reset": _("Change password"),
            "profile": _("Profile"),
        },
        "search": {
            "delete_confirm": _("Are you sure you want to delete selected entries?"),
            "deleted": _("${count} item(s) deleted"),
            "disable_confirm": _("Are you sure you want to disable selected entries?"),
            "disabled": _("${count} item(s) disabled"),
            "enable_confirm": _("Are you sure you want to enable selected entries?"),
            "enabled": _("${count} item(s) enabled"),
            "partial_delete": _("Some entries were not deleted"),
            "placeholder": _("Search"),
            "placeholder_filter": _("Filter"),
            "quick_links": _("Quick Links"),
            "select_all": _("Select All"),
            "truncated": _("Query returned more results than the configured size limit. Displaying the first ${counter} results."),
            "unselect_all": _("Unselect All"),
        },
        "ssbrowser-page": {
            "header": _(
                "<h1>Browser Kerberos Setup</h1>\n"
                "\n"
            ),
            "firefox-header": _(
                "<h2>Firefox</h2>\n"
                "\n"
                "<p>\n"
                "            You can configure Firefox to use Kerberos for "
                "Single Sign-on. The following instructions will guide you in "
                "configuring your web browser to send your Kerberos "
                "credentials to the appropriate Key Distribution Center which "
                "enables Single Sign-on.\n"
                "</p>\n"
                "\n"
            ),
            "firefox-actions": _(
                "<ol>\n"
                "<li>\n"
                "<p>\n"
                "<a href=\"ca.crt\" id=\"ca-link\" class=\"btn btn-default\">"
                "Import Certificate Authority certificate</a>\n"
                "</p>\n"
                "<p>\n"
                "                    Make sure you select <b>all three</b> "
                "checkboxes.\n"
                "</p>\n"
                "</li>\n"
                "<li>\n"
                "                In the address bar of Firefox, type <code>"
                "about:config</code> to display the list of current "
                "configuration options.\n"
                "</li>\n"
                "<li>\n"
                "                In the Filter field, type <code>negotiate"
                "</code> to restrict the list of options.\n"
                "</li>\n"
                "<li>\n"
                "                Double-click the <code>network.negotiate-auth"
                ".trusted-uris</code> entry to display the Enter string value "
                "dialog box.\n"
                "</li>\n"
                "<li>\n"
                "                Enter the name of the domain against which "
                "you want to authenticate, for example, <code class=\""
                "example-domain\">.example.com.</code>\n"
                "</li>\n"
                "<li><a href=\"../ui/index.html\" id=\"return-link\" class=\""
                "btn btn-default\">Return to Web UI</a></li>\n"
                "</ol>\n"
                "\n"
            ),
            "chrome-header": _(
                "<h2>Chrome</h2>\n"
                "\n"
                "<p>\n"
                "            You can configure Chrome to use Kerberos for "
                "Single Sign-on. The following instructions will guide you in "
                "configuring your web browser to send your Kerberos "
                "credentials to the appropriate Key Distribution Center which "
                "enables Single Sign-on.\n"
                "</p>\n"
                "\n"
            ),
            "chrome-certificate": _(
                "<h3>Import CA Certificate</h3>\n"
                "<ol>\n"
                "<li>\n"
                "                Download the <a href=\"ca.crt\">CA "
                "certificate</a>. Alternatively, if the host is also an IdM "
                "client, you can find the certificate in /etc/ipa/ca.crt.\n"
                "</li>\n"
                "<li>\n"
                "                Click the menu button with the <em>Customize "
                "and control Google Chrome</em> tooltip, which is by default "
                "in the top right-hand corner of Chrome, and click <em>"
                "Settings</em>.\n"
                "</li>\n"
                "<li>\n"
                "                Click <em>Show advanced settings</em> to "
                "display more options, and then click the <em>Manage "
                "certificates</em> button located under the HTTPS/SSL heading."
                "\n"
                "</li>\n"
                "<li>\n"
                "                In the <em>Authorities</em> tab, click the "
                "<em>Import</em> button at the bottom.\n"
                "</li>\n"
                "<li>Select the CA certificate file that you downloaded in the"
                " first step.</li>\n"
                "</ol>\n"
                "\n"
            ),
            "chrome-spnego": _(
                "<h3>\n"
                "            Enable SPNEGO (Simple and Protected GSSAPI "
                "Negotiation Mechanism) to Use Kerberos Authentication\n"
                "            in Chrome\n"
                "</h3>\n"
                "<ol>\n"
                "<li>\n"
                "                Make sure you have the necessary directory "
                "created by running:\n"
                "<div><code>\n"
                "                    [root@client]# mkdir -p /etc/opt/chrome/"
                "policies/managed/\n"
                "</code></div>\n"
                "</li>\n"
                "<li>\n"
                "                Create a new <code>/etc/opt/chrome/policies/"
                "managed/mydomain.json</code> file with write privileges "
                "limited to the system administrator or root, and include the "
                "following line:\n"
                "<div><code>\n"
                "                    { \"AuthServerWhitelist\": \"*<span "
                "class=\"example-domain\">.example.com.</span>\" }\n"
                "</code></div>\n"
                "<div>\n"
                "                    You can do this by running:\n"
                "</div>\n"
                "<div><code>\n"
                "                    [root@server]# echo \'{ \""
                "AuthServerWhitelist\": \"*<span class=\"example-domain\">"
                ".example.com.</span>\" }' > /etc/opt/chrome/policies/managed/"
                "mydomain.json\n"
                "</code></div>\n"
                "</li>\n"
                "</ol>\n"
                "<ol>\n"
                "<p>\n"
                "<strong>Note:</strong> If using Chromium, use <code>/etc/"
                "chromium/policies/managed/</code> instead of <code>/etc/opt/"
                "chrome/policies/managed/</code> for the two SPNEGO Chrome "
                "configuration steps above.\n"
                "</p>\n"
                "</ol>\n"
                "\n"
            ),
            "ie-header": _(
                "<h2>Internet Explorer</h2>\n"
                "<p><strong>WARNING:</strong> Internet Explorer is no longer a"
                " supported browser.</p>\n"
                "<p>\n"
                "            Once you are able to log into the workstation "
                "with your kerberos key you are now able to use that ticket in"
                " Internet Explorer.\n"
                "</p>\n"
                "<p>\n"
            ),
            "ie-actions": _(
                "<strong>Log into the Windows machine using an account of your"
                " Kerberos realm (administrative domain)</strong>\n"
                "</p>\n"
                "<p>\n"
                "<strong>In Internet Explorer, click Tools, and then click "
                "Internet Options.</strong>\n"
                "</p>\n"
                "<div>\n"
                "<ol>\n"
                "<li>Click the Security tab</li>\n"
                "<li>Click Local intranet</li>\n"
                "<li>Click Sites </li>\n"
                "<li>Click Advanced </li>\n"
                "<li>Add your domain to the list</li>\n"
                "</ol>\n"
                "<ol>\n"
                "<li>Click the Security tab</li>\n"
                "<li>Click Local intranet</li>\n"
                "<li>Click Custom Level</li>\n"
                "<li>Select Automatic logon only in Intranet zone</li>\n"
                "</ol>\n"
                "\n"
                "<ol>\n"
                "<li> Visit a kerberized web site using IE (You must use the "
                "fully-qualified Domain Name in the URL)</li>\n"
                "<li><strong> You are all set.</strong></li>\n"
                "</ol>\n"
                "</div>\n"
                "\n"
            ),
        },
        "status": {
            "disable": _("Disable"),
            "disabled": _("Disabled"),
            "enable": _("Enable"),
            "enabled": _("Enabled"),
            "label": _("Status"),
            "working": _("Working"),
        },
        "tabs": {
            "audit": _("Audit"),
            "authentication": _("Authentication"),
            "automember": _("Automember"),
            "automount": _("Automount"),
            "cert": _("Certificates"),
            "dns": _("DNS"),
            "hbac": _("Host-Based Access Control"),
            "identity": _("Identity"),
            "ipaserver": _("IPA Server"),
            "network_services": _("Network Services"),
            "policy": _("Policy"),
            "role": _("Role-Based Access Control"),
            "sudo": _("Sudo"),
            "topology": _("Topology"),
            "trust": _("Trusts"),
        },
        "true": _("True"),
        "unauthorized-page": _(
            "<h1>Unable to verify your Kerberos credentials</h1>\n"
            "<p>\n"
            "            Please make sure that you have valid Kerberos "
            "tickets (obtainable via <strong>kinit</strong>), and that you"
            " have configured your browser correctly.\n"
            "</p>\n"
            "\n"
            "<h2>Browser configuration</h2>\n"
            "\n"
            "<div id=\"first-time\">\n"
            "<p>\n"
            "                If this is your first time, please <a href="
            "\"ssbrowser.html\">configure your browser</a>.\n"
            "</p>\n"
            "</div>\n"
        ),
        "widget": {
            "api_browser": _("API Browser"),
            "first": _("First"),
            "last": _("Last"),
            "next": _("Next"),
            "page": _("Page"),
            "prev": _("Prev"),
            "undo": _("Undo"),
            "undo_title": _("Undo this change."),
            "undo_all": _("Undo All"),
            "undo_all_title": _("Undo all changes in this field."),
            "validation": {
                "error": _("Text does not match field pattern"),
                "datetime": _("Must be an UTC date/time value (e.g., \"2014-01-20 17:58:01Z\")"),
                "decimal": _("Must be a decimal number"),
                "format": _("Format error"),
                "integer": _("Must be an integer"),
                "ip_address": _('Not a valid IP address'),
                "ip_v4_address": _('Not a valid IPv4 address'),
                "ip_v6_address": _('Not a valid IPv6 address'),
                "max_value": _("Maximum value is ${value}"),
                "min_value": _("Minimum value is ${value}"),
                "net_address": _("Not a valid network address (examples: 2001:db8::/64, 192.0.2.0/24)"),
                "parse": _("Parse error"),
                "positive_number": _("Must be a positive number"),
                "port": _("'${port}' is not a valid port"),
                "required": _("Required field"),
                "unsupported": _("Unsupported value"),
            },
        },
    }
    has_output = (
        Output('texts', dict, doc=_('Dict of I18N messages')),
    )
    def execute(self, **options):
        return dict(texts=json_serialize(self.messages))



#!/usr/bin/python2

"""Copy the IPA schema to the CA directory server instance

You need to run this script to prepare a 2.2 or 3.0 IPA master for
installation of a 3.1 replica.

Once a 3.1 replica is in the domain, every older CA master will emit schema
replication errors until this script is run on it.

"""

# DO NOT TOUCH THIS CODE, IT MUST BE COMPATIBLE WITH RHEL6
# disable pylint because current codebase didn't match RHEL6 code
# pylint: disable=all

import os
import sys
import pwd
import shutil

from hashlib import sha1

from ipapython import ipautil
from ipapython.ipa_log_manager import root_logger, standard_logging_setup
from ipaserver.install.dsinstance import schema_dirname
from ipalib import api

# oh dear, this is an old IPA (3.0+)
from ipaserver.install.dsinstance import DS_USER
from ipaserver.install.cainstance import PKI_USER
from ipapython import services

# for mod_nss
from ipaserver.install.httpinstance import NSS_CONF
from ipaserver.install.httpinstance import HTTPInstance
from ipaserver.install import installutils
from ipapython import sysrestore

SERVERID = "PKI-IPA"
SCHEMA_FILENAMES = (
    "60kerberos.ldif",
    "60samba.ldif",
    "60ipaconfig.ldif",
    "60basev2.ldif",
    "60basev3.ldif",
    "60ipadns.ldif",
    "61kerberos-ipav3.ldif",
    "65ipacertstore.ldif",
    "65ipasudo.ldif",
    "70ipaotp.ldif",
    "05rfc2247.ldif",
)


def _sha1_file(filename):
    with open(filename, 'rb') as f:
        return sha1(f.read()).hexdigest()


def add_ca_schema():
    """Copy IPA schema files into the CA DS instance
    """
    pki_pent = pwd.getpwnam(PKI_USER)
    ds_pent = pwd.getpwnam(DS_USER)
    for schema_fname in SCHEMA_FILENAMES:
        source_fname = os.path.join(ipautil.SHARE_DIR, schema_fname)
        target_fname = os.path.join(schema_dirname(SERVERID), schema_fname)
        if not os.path.exists(source_fname):
            root_logger.debug('File does not exist: %s', source_fname)
            continue
        if os.path.exists(target_fname):
            target_sha1 = _sha1_file(target_fname)
            source_sha1 = _sha1_file(source_fname)
            if target_sha1 != source_sha1:
                target_size = os.stat(target_fname).st_size
                source_size = os.stat(source_fname).st_size
                root_logger.info('Target file %s exists but the content is '
                                 'different', target_fname)
                root_logger.info('\tTarget file: sha1: %s, size: %s B',
                                 target_sha1, target_size)
                root_logger.info('\tSource file: sha1: %s, size: %s B',
                                 source_sha1, source_size)
                if not ipautil.user_input("Do you want replace %s file?" %
                                          target_fname, True):
                    continue

            else:
                root_logger.info(
                    'Target exists, not overwriting: %s', target_fname)
                continue
        try:
            shutil.copyfile(source_fname, target_fname)
        except IOError as e:
            root_logger.warning('Could not install %s: %s', target_fname, e)
        else:
            root_logger.info('Installed %s', target_fname)
        os.chmod(target_fname, 0o440)    # read access for dirsrv user/group
        os.chown(target_fname, pki_pent.pw_uid, ds_pent.pw_gid)


def restart_pki_ds():
    """Restart the CA DS instance to pick up schema changes
    """
    root_logger.info('Restarting CA DS')
    services.service('dirsrv').restart(SERVERID)


# The ipa-3-0 set_directive() has very loose comparision of directive
# which would cause multiple NSSCipherSuite to be added so provide
# a custom function for it.
def set_directive(filename, directive, value, quotes=True, separator=' '):
    """Set a name/value pair directive in a configuration file.

       A value of None means to drop the directive.

       This has only been tested with nss.conf
    """
    valueset = False
    st = os.stat(filename)
    fd = open(filename)
    newfile = []
    for line in fd:
        if line.lstrip().startswith(directive):
            valueset = True
            if value is not None:
                if quotes:
                    newfile.append('%s%s"%s"\n' %
                                   (directive, separator, value))
                else:
                    newfile.append('%s%s%s\n' % (directive, separator, value))
        else:
            newfile.append(line)
    fd.close()
    if not valueset:
        if value is not None:
            if quotes:
                newfile.append('%s%s"%s"\n' % (directive, separator, value))
            else:
                newfile.append('%s%s%s\n' % (directive, separator, value))

    fd = open(filename, "w")
    fd.write("".join(newfile))
    fd.close()
    os.chown(filename, st.st_uid, st.st_gid)  # reset perms


def update_mod_nss_cipher_suite():
    add_ciphers = ['ecdhe_rsa_aes_128_sha', 'ecdhe_rsa_aes_256_sha']
    ciphers = installutils.get_directive(NSS_CONF, 'NSSCipherSuite')

    # Run through once to see if any of the new ciphers are there but
    # disabled. If they are then enable them.
    lciphers = ciphers.split(',')
    new_ciphers = []
    for cipher in lciphers:
        for add in add_ciphers:
            if cipher.endswith(add):
                if cipher.startswith('-'):
                    cipher = '+%s' % add
        new_ciphers.append(cipher)

    # Run through again and add remaining ciphers as enabled.
    for add in add_ciphers:
        if add not in ciphers:
            new_ciphers.append('+%s' % add)

    ciphers = ','.join(new_ciphers)
    set_directive(NSS_CONF, 'NSSCipherSuite', ciphers, False)
    root_logger.info('Updated Apache cipher list')


def restart_http():
    root_logger.info('Restarting HTTP')
    fstore = sysrestore.FileStore('/var/lib/ipa/sysrestore')
    http = HTTPInstance(fstore)
    http.restart()


def main():
    if os.getegid() != 0:
        sys.exit("Must be root to run this script")
    standard_logging_setup(verbose=True)

    # In 3.0, restarting needs access to api.env
    api.bootstrap_with_global_options(context='server')

    add_ca_schema()
    restart_pki_ds()
    update_mod_nss_cipher_suite()
    restart_http()

    root_logger.info('Schema updated successfully')


if __name__ == '__main__':
    main()



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

from collections import namedtuple

from ipalib import _
from ipalib import Command
from ipalib import errors
from ipalib import output
from ipalib.parameters import Int
from ipalib.plugable import Registry
from ipalib.constants import DOMAIN_LEVEL_0, MIN_DOMAIN_LEVEL

from ipapython.dn import DN


__doc__ = _("""
Raise the IPA Domain Level.
""")

register = Registry()

DomainLevelRange = namedtuple('DomainLevelRange', ['min', 'max'])

domainlevel_output = (
    output.Output('result', int, _('Current domain level:')),
)


def get_domainlevel_dn(api):
    domainlevel_dn = DN(
        ('cn', 'Domain Level'),
        ('cn', 'ipa'),
        ('cn', 'etc'),
        api.env.basedn
    )

    return domainlevel_dn


def get_domainlevel_range(master_entry):
    try:
        return DomainLevelRange(
            int(master_entry['ipaMinDomainLevel'][0]),
            int(master_entry['ipaMaxDomainLevel'][0])
        )
    except KeyError:
        return DomainLevelRange(DOMAIN_LEVEL_0, DOMAIN_LEVEL_0)


def check_conflict_entries(ldap, api, desired_value):
    """
    Check if conflict entries exist in topology subtree
    """

    container_dn = DN(
        ('cn', 'ipa'),
        ('cn', 'etc'),
        api.env.basedn
    )
    conflict = "(nsds5replconflict=*)"
    subentry = "(|(objectclass=ldapsubentry)(objectclass=*))"
    try:
        ldap.get_entries(
            filter="(& %s %s)" % (conflict, subentry),
            base_dn=container_dn,
            scope=ldap.SCOPE_SUBTREE)
        message = _("Domain Level cannot be raised to {0}, "
                    "existing replication conflicts have to be resolved."
                    ).format(desired_value)
        raise errors.InvalidDomainLevelError(reason=message)
    except errors.NotFound:
        pass


def get_master_entries(ldap, api):
    """
    Returns list of LDAPEntries representing IPA masters.
    """
    dn = DN(api.env.container_masters, api.env.basedn)
    masters, _dummy = ldap.find_entries(
        filter="(cn=*)",
        base_dn=dn,
        scope=ldap.SCOPE_ONELEVEL,
        paged_search=True,  # we need to make sure to get all of them
    )
    return masters


@register()
class domainlevel_get(Command):
    __doc__ = _('Query current Domain Level.')

    has_output = domainlevel_output

    def execute(self, *args, **options):
        ldap = self.api.Backend.ldap2
        entry = ldap.get_entries(
            get_domainlevel_dn(self.api),
            scope=ldap.SCOPE_BASE,
            filter='(objectclass=ipadomainlevelconfig)',
            attrs_list=['ipaDomainLevel']
        )[0]

        try:
            value = int(entry.single_value['ipaDomainLevel'])
            return {'result': value}
        except KeyError:
            raise errors.NotFound(
                reason=_(
                    'Server does not support domain level functionality'))


@register()
class domainlevel_set(Command):
    __doc__ = _('Change current Domain Level.')

    has_output = domainlevel_output

    takes_args = (
        Int('ipadomainlevel',
            cli_name='level',
            label=_('Domain Level'),
            minvalue=MIN_DOMAIN_LEVEL,
        ),
    )

    def execute(self, *args, **options):
        """
        Checks all the IPA masters for supported domain level ranges.

        If the desired domain level is within the supported range of all
        masters, it will be raised.

        Domain level cannot be lowered.
        """

        ldap = self.api.Backend.ldap2

        current_entry = ldap.get_entry(get_domainlevel_dn(self.api))
        current_value = int(current_entry.single_value['ipadomainlevel'])
        desired_value = int(args[0])

        # Domain level cannot be lowered
        if int(desired_value) < int(current_value):
            message = _("Domain Level cannot be lowered.")
            raise errors.InvalidDomainLevelError(reason=message)

        # Check if every master supports the desired level
        for master in get_master_entries(ldap, self.api):
            supported = get_domainlevel_range(master)

            if supported.min > desired_value or supported.max < desired_value:
                message = _("Domain Level cannot be raised to {0}, server {1} "
                            "does not support it."
                            ).format(desired_value, master['cn'][0])
                raise errors.InvalidDomainLevelError(reason=message)

        # Check if conflict entries exist in topology subtree
        # should be resolved first
        check_conflict_entries(ldap, self.api, desired_value)

        current_entry.single_value['ipaDomainLevel'] = desired_value
        ldap.update_entry(current_entry)

        return {'result': int(current_entry.single_value['ipaDomainLevel'])}



# Authors:
#   Jr Aquino <jr.aquino@citrixonline.com>
#
# Copyright (C) 2010  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ipalib import api, errors
from ipalib import Str
from ipalib.plugable import Registry
from .baseldap import (
    LDAPObject,
    LDAPCreate,
    LDAPDelete,
    LDAPUpdate,
    LDAPSearch,
    LDAPRetrieve)
from ipalib import _, ngettext
from ipapython.dn import DN

__doc__ = _("""
Sudo Commands

Commands used as building blocks for sudo

EXAMPLES:

 Create a new command
   ipa sudocmd-add --desc='For reading log files' /usr/bin/less

 Remove a command
   ipa sudocmd-del /usr/bin/less

""")

register = Registry()

topic = 'sudo'

@register()
class sudocmd(LDAPObject):
    """
    Sudo Command object.
    """
    container_dn = api.env.container_sudocmd
    object_name = _('sudo command')
    object_name_plural = _('sudo commands')
    object_class = ['ipaobject', 'ipasudocmd']
    permission_filter_objectclasses = ['ipasudocmd']
    # object_class_config = 'ipahostobjectclasses'
    search_attributes = [
        'sudocmd', 'description',
    ]
    default_attributes = [
        'sudocmd', 'description', 'memberof',
    ]
    attribute_members = {
        'memberof': ['sudocmdgroup'],
    }
    uuid_attribute = 'ipauniqueid'
    rdn_attribute = 'ipauniqueid'
    managed_permissions = {
        'System: Read Sudo Commands': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'description', 'ipauniqueid', 'memberof', 'objectclass',
                'sudocmd',
            },
        },
        'System: Add Sudo Command': {
            'ipapermright': {'add'},
            'replaces': [
                '(target = "ldap:///sudocmd=*,cn=sudocmds,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Add Sudo command";allow (add) groupdn = "ldap:///cn=Add Sudo command,cn=permissions,cn=pbac,$SUFFIX";)',
                '(targetfilter = "(objectclass=ipasudocmd)")(target = "ldap:///cn=sudocmds,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Add Sudo command";allow (add) groupdn = "ldap:///cn=Add Sudo command,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Sudo Administrator'},
        },
        'System: Delete Sudo Command': {
            'ipapermright': {'delete'},
            'replaces': [
                '(target = "ldap:///sudocmd=*,cn=sudocmds,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Delete Sudo command";allow (delete) groupdn = "ldap:///cn=Delete Sudo command,cn=permissions,cn=pbac,$SUFFIX";)',
                '(targetfilter = "(objectclass=ipasudocmd)")(target = "ldap:///cn=sudocmds,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Delete Sudo command";allow (delete) groupdn = "ldap:///cn=Delete Sudo command,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Sudo Administrator'},
        },
        'System: Modify Sudo Command': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {'description'},
            'replaces': [
                '(targetattr = "description")(target = "ldap:///sudocmd=*,cn=sudocmds,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Modify Sudo command";allow (write) groupdn = "ldap:///cn=Modify Sudo command,cn=permissions,cn=pbac,$SUFFIX";)',
                '(targetfilter = "(objectclass=ipasudocmd)")(targetattr = "description")(target = "ldap:///cn=sudocmds,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Modify Sudo command";allow (write) groupdn = "ldap:///cn=Modify Sudo command,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Sudo Administrator'},
        },
    }

    label = _('Sudo Commands')
    label_singular = _('Sudo Command')

    takes_params = (
        Str('sudocmd',
            cli_name='command',
            label=_('Sudo Command'),
            primary_key=True,
        ),
        Str('description?',
            cli_name='desc',
            label=_('Description'),
            doc=_('A description of this command'),
        ),
    )

    def get_dn(self, *keys, **options):
        if keys[-1].endswith('.'):
            keys = (keys[:-1] + (keys[-1][:-1], ))
        dn = super(sudocmd, self).get_dn(*keys, **options)
        try:
            self.backend.get_entry(dn, [''])
        except errors.NotFound:
            try:
                entry_attrs = self.backend.find_entry_by_attr(
                    'sudocmd', keys[-1], self.object_class, [''],
                    DN(self.container_dn, api.env.basedn))
                dn = entry_attrs.dn
            except errors.NotFound:
                pass
        return dn


@register()
class sudocmd_add(LDAPCreate):
    __doc__ = _('Create new Sudo Command.')

    msg_summary = _('Added Sudo Command "%(value)s"')


@register()
class sudocmd_del(LDAPDelete):
    __doc__ = _('Delete Sudo Command.')

    msg_summary = _('Deleted Sudo Command "%(value)s"')

    def pre_callback(self, ldap, dn, *keys, **options):
        filters = [
            ldap.make_filter_from_attr(attr, dn)
            for attr in ('memberallowcmd', 'memberdenycmd')]
        filter = ldap.combine_filters(filters, ldap.MATCH_ANY)
        filter = ldap.combine_filters(
            (filter, ldap.make_filter_from_attr('objectClass', 'ipasudorule')),
            ldap.MATCH_ALL)
        dependent_sudorules = []
        try:
            entries, _truncated = ldap.find_entries(
                filter, ['cn'],
                base_dn=DN(api.env.container_sudorule, api.env.basedn))
        except errors.NotFound:
            pass
        else:
            for entry_attrs in entries:
                [cn] = entry_attrs['cn']
                dependent_sudorules.append(cn)

        if dependent_sudorules:
            raise errors.DependentEntry(
                key=keys[0], label='sudorule',
                dependent=', '.join(dependent_sudorules))
        return dn


@register()
class sudocmd_mod(LDAPUpdate):
    __doc__ = _('Modify Sudo Command.')

    msg_summary = _('Modified Sudo Command "%(value)s"')


@register()
class sudocmd_find(LDAPSearch):
    __doc__ = _('Search for Sudo Commands.')

    msg_summary = ngettext(
        '%(count)d Sudo Command matched', '%(count)d Sudo Commands matched', 0
    )


@register()
class sudocmd_show(LDAPRetrieve):
    __doc__ = _('Display Sudo Command.')



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#

"""This module provides tests for NFS-related features like
   krb5 NFS and automount locations.

   Wishlist
   * add automount direct and indirect maps
   * add automount /home for the "seattle" location only
   * validate it is not available in another location
   * krb5 /home for IdM users in test_automount
   * store nfs configuration in a single place
"""

from __future__ import absolute_import

import time
import re

from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration import tasks
from ipaplatform.paths import paths

import os

# give some time for units to stabilize
# otherwise we get transient errors
WAIT_AFTER_INSTALL = 5
WAIT_AFTER_UNINSTALL = WAIT_AFTER_INSTALL


class TestInit(IntegrationTest):

    @classmethod
    def fix_resolv_conf(cls, client, server):

        contents = client.get_file_contents(paths.RESOLV_CONF,
                                            encoding='utf-8')
        nameserver = 'nameserver %s\n' % server.ip
        if not contents.startswith(nameserver):
            contents = nameserver + contents.replace(nameserver, '')
            client.run_command([
                '/usr/bin/cp', paths.RESOLV_CONF,
                '%s.sav' % paths.RESOLV_CONF
            ])
            client.put_file_contents(paths.RESOLV_CONF, contents)

    @classmethod
    def restore_resolv_conf(cls, client):
        client.run_command([
            '/usr/bin/cp',
            '%s.sav' % paths.RESOLV_CONF,
            paths.RESOLV_CONF
        ])


class TestNFS(TestInit):

    num_replicas = 2
    num_clients = 1
    topology = 'star'

    @classmethod
    def install(cls, mh):

        tasks.install_master(cls.master, setup_dns=True)
        clients = (cls.clients[0], cls.replicas[0], cls.replicas[1])
        for client in clients:
            cls.fix_resolv_conf(client, cls.master)
            tasks.install_client(cls.master, client)
            client.run_command(["cat", "/etc/resolv.conf"])

    def cleanup(self):

        nfssrv = self.clients[0]
        nfsclt = self.replicas[0]
        automntclt = self.replicas[1]

        nfsclt.run_command(["umount", "-a", "-t", "nfs4"])
        nfsclt.run_command(["systemctl", "stop", "rpc-gssd"])

        nfssrv.run_command(["systemctl", "stop", "nfs-server"])
        nfssrv.run_command(["systemctl", "disable", "nfs-server"])
        nfssrv.run_command([
            "rm", "-f", "/etc/exports.d/krbnfs.exports",
            "/etc/exports.d/stdnfs.exports"
        ])

        nfssrv.run_command(["rm", "-rf", "/exports"])

        tasks.uninstall_client(nfsclt)
        tasks.uninstall_client(nfssrv)
        self.master.run_command([
            "ipa", "host-mod", automntclt.hostname,
            "--location", "''"
        ])
        # not strictly necessary, but this exercises automountlocation-del
        self.master.run_command([
            "ipa", "automountlocation-del", "seattle"
        ])
        nfsclt.run_command(["systemctl", "restart", "nfs-utils"])
        nfssrv.run_command(["systemctl", "restart", "nfs-utils"])
        for client in (nfssrv, nfsclt, automntclt):
            self.restore_resolv_conf(client)
        tasks.uninstall_master(self.master)

    def test_prepare_users(self):

        users = {
            "athena": "p",
            "euripides": "s"
        }
        temp_pass = 'temppass'
        for user in users:
            self.master.run_command([
                "ipa", "user-add",
                "%s" % user, "--first", "%s" % user,
                "--last", "%s" % users[user],
                '--password'], stdin_text="%s\n%s\n" % (temp_pass, temp_pass)
            )
            self.master.run_command(["kdestroy", "-A"])
            password = "Secret123"
            user_kinit = "%s\n%s\n%s\n" % (temp_pass, password, password)
            self.master.run_command(
                ['kinit', user], stdin_text=user_kinit
            )
            self.master.run_command(["kdestroy", "-A"])
            tasks.kinit_admin(self.master)

    def test_krb5_nfsd(self):

        nfssrv = self.clients[0]

        # NFS keytab management
        self.master.run_command([
            "ipa", "service-add", "nfs/%s" % nfssrv.hostname
        ])
        nfssrv.run_command([
            "ipa-getkeytab", "-p", "nfs/%s" % nfssrv.hostname,
            "-k", "/etc/krb5.keytab"
        ])
        nfssrv.run_command(["systemctl", "restart", "nfs-server"])
        nfssrv.run_command(["systemctl", "enable", "nfs-server"])
        time.sleep(WAIT_AFTER_INSTALL)

        basedir = "exports"
        exports = {
            "krbnfs": "*(sec=krb5p,rw)",
            "stdnfs": "*(ro)",
            "home": "*(sec=krb5p,rw)"
        }
        for export in exports:
            exportpath = os.sep.join(('', basedir, export))
            exportfile = os.sep.join((
                '', 'etc', 'exports.d', "%s.exports" % export
            ))
            exportline = " ".join((exportpath, exports[export]))
            nfssrv.run_command(["mkdir", "-p", exportpath])
            nfssrv.run_command(["chmod", "770", exportpath])
            nfssrv.put_file_contents(exportfile, exportline)
            nfssrv.run_command(["cat", exportfile])
        nfssrv.run_command(["exportfs", "-r"])
        nfssrv.run_command(["exportfs", "-s"])

    def test_krb5_nfs_manual_configuration(self):

        nfssrv = self.clients[0]
        nfsclt = self.replicas[0]

        nfsclt.run_command(["systemctl", "restart", "rpc-gssd"])
        time.sleep(WAIT_AFTER_INSTALL)
        mountpoints = ("/mnt/krb", "/mnt/std", "/home")
        for mountpoint in mountpoints:
            nfsclt.run_command(["mkdir", "-p", mountpoint])
        nfsclt.run_command([
            "systemctl", "status", "gssproxy"
        ])
        nfsclt.run_command([
            "systemctl", "status", "rpc-gssd"
        ])
        nfsclt.run_command([
            "mount", "-t", "nfs4", "-o", "sec=krb5p,vers=4.0",
            "%s:/exports/krbnfs" % nfssrv.hostname, "/mnt/krb", "-v"
        ])
        nfsclt.run_command([
            "mount", "-t", "nfs4", "-o", "sec=krb5p,vers=4.0",
            "%s:/exports/home" % nfssrv.hostname, "/home", "-v"
        ])

    def test_automount(self):
        """
        Test if ipa-client-automount behaves as expected
        """

        nfssrv = self.clients[0]
        automntclt = self.replicas[1]

        self.master.run_command([
            "ipa", "automountlocation-add", "seattle"
        ])
        self.master.run_command([
            "ipa", "automountmap-add", "seattle", "auto.home"
        ])
        self.master.run_command([
            "ipa", "automountkey-add", "seattle", "auto.home",
            "--key='*'", "--info=sec=krb5p,vers=4"
            " 'rhel8-nfsserver0.laptop.example.org:/export/home/&'"
        ])
        self.master.run_command([
            "ipa", "automountkey-add", "seattle", "auto.master",
            "--key=/home", "--info=auto.home"
        ])

        self.master.run_command([
            "ipa", "host-mod", automntclt.hostname,
            "--location", "seattle"
        ])

        # systemctl non-fatal errors will only be displayed
        # if ipa-client-automount is launched with --debug
        result1 = automntclt.run_command([
            'ipa-client-automount', '--location', 'seattle',
            '-U', '--debug'
        ])

        time.sleep(WAIT_AFTER_INSTALL)

        # systemctl non-fatal errors will show up like this:
        # stderr=Failed to restart nfs-secure.service: \
        #        Unit nfs-secure.service not found.
        # normal output:
        # stderr=
        m1 = re.search(r'(?<=stderr\=Failed).+', result1.stderr_text)
        # maybe re-use m1.group(0) if it exists.
        assert m1 is None

        # https://pagure.io/freeipa/issue/7918
        # check whether idmapd.conf was setup using the IPA domain
        automntclt.run_command([
            "grep", "Domain = %s" % self.master.domain.name, "/etc/idmapd.conf"
        ])

        automntclt.run_command([
            "mount", "-t", "nfs4", "-o", "sec=krb5p,vers=4.0",
            "%s:/exports/home" % nfssrv.hostname, "/home", "-v"
        ])

        # TODO leverage users

        automntclt.run_command(["umount", "-a", "-t", "nfs4"])

        result2 = automntclt.run_command([
            'ipa-client-automount', '--uninstall', '-U', '--debug'
        ])
        m2 = re.search(r'(?<=stderr\=Failed).+', result2.stderr_text)
        assert m2 is None

        time.sleep(WAIT_AFTER_UNINSTALL)

        # https://pagure.io/freeipa/issue/7918
        # test for --idmap-domain DNS
        automntclt.run_command([
            'ipa-client-automount', '--location', 'default',
            '-U', '--debug', "--idmap-domain", "DNS"
        ])

        time.sleep(WAIT_AFTER_INSTALL)

        # check whether idmapd.conf was setup properly:
        # grep must not find any configured Domain.
        result = automntclt.run_command(
            ["grep", "^Domain =", "/etc/idmapd.conf"], raiseonerr=False
        )
        assert result.returncode == 1

        automntclt.run_command([
            'ipa-client-automount', '--uninstall', '-U', '--debug'
        ])

        time.sleep(WAIT_AFTER_UNINSTALL)

        # https://pagure.io/freeipa/issue/7918
        # test for --idmap-domain exampledomain.net
        nfs_domain = "exampledomain.net"
        automntclt.run_command([
            'ipa-client-automount', '--location', 'default',
            '-U', '--debug', "--idmap-domain", nfs_domain
        ])
        # check whether idmapd.conf was setup using nfs_domain
        automntclt.run_command([
            "grep", "Domain = %s" % nfs_domain, "/etc/idmapd.conf"
        ])

        time.sleep(WAIT_AFTER_INSTALL)

        automntclt.run_command([
            'ipa-client-automount', '--uninstall', '-U', '--debug'
        ])

        time.sleep(WAIT_AFTER_UNINSTALL)

        self.cleanup()



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#

from __future__ import absolute_import

import paramiko
import pytest

from ipaplatform.osinfo import osinfo
from ipaplatform.paths import paths
from ipaplatform.tasks import tasks as platformtasks
from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration import tasks

class TestUserPermissions(IntegrationTest):
    topology = 'star'
    altadmin = "altadmin"

    @classmethod
    def install(cls, mh):
        super(TestUserPermissions, cls).install(mh)
        tasks.kinit_admin(cls.master)

        # Create a new user altadmin
        password_confirmation = "%s\n%s\n" % (cls.master.config.admin_password,
                                              cls.master.config.admin_password)
        cls.master.run_command(['ipa', 'user-add', cls.altadmin,
                                '--first', cls.altadmin,
                                '--last', cls.altadmin,
                                '--password'],
                               stdin_text=password_confirmation)

        # Add altadmin to the group cn=admins
        cls.master.run_command(['ipa', 'group-add-member', 'admins',
                                '--users', cls.altadmin])

        # kinit as altadmin to initialize the password
        altadmin_kinit = "%s\n%s\n%s\n" % (cls.master.config.admin_password,
                                           cls.master.config.admin_password,
                                           cls.master.config.admin_password)
        cls.master.run_command(['kinit', cls.altadmin],
                               stdin_text=altadmin_kinit)
        cls.master.run_command(['kdestroy', '-A'])

    def test_delete_preserve_as_alternate_admin(self):
        """
        Test that a user member of admins group can call delete --preserve.

        This is a test case for issue 7342
        """

        # kinit admin
        tasks.kinit_admin(self.master)

        # Create a new user 'testuser' with a password
        testuser = 'testuser'
        password = 'Secret123'
        testuser_password_confirmation = "%s\n%s\n" % (password,
                                                       password)
        self.master.run_command(['ipa', 'user-add', testuser,
                                 '--first', testuser,
                                 '--last', testuser,
                                 '--password'],
                                stdin_text=testuser_password_confirmation)

        # kinit as altadmin
        self.master.run_command(['kinit', self.altadmin],
                                stdin_text=self.master.config.admin_password)

        # call ipa user-del --preserve
        self.master.run_command(['ipa', 'user-del', '--preserve', testuser])

    @pytest.mark.skipif(
        not platformtasks.is_selinux_enabled(),
        reason="Test needs SELinux enabled")
    @pytest.mark.xfail(
        osinfo.id == 'fedora' and osinfo.version_number <= (28,),
        reason='sssd ticket 3819', strict=True)
    def test_selinux_user_optimized(self):
        """
        Check that SELinux login context is set on first login for the
        user, even if the user is not mapped to a specific SELinux user.

        Related ticket https://pagure.io/SSSD/sssd/issue/3819.
        """
        # Scenario: add an IPA user with non-default home dir, login through
        # ssh as this user and check that there is a SELinux user mapping
        # for the user with `semanage login -l`.

        # kinit admin
        tasks.kinit_admin(self.master)

        testuser = 'testuser_selinux'
        password = 'Secret123'
        testuser_password_confirmation = "%s\n%s\n" % (password,
                                                       password)
        self.master.run_command(['ipa', 'user-add', testuser,
                                 '--first', testuser,
                                 '--last', testuser,
                                 '--password',
                                 '--homedir',
                                 '/root/{}'.format(testuser)],
                                stdin_text=testuser_password_confirmation)

        # login to the system
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(self.master.hostname,
                       username=testuser,
                       password=password)
        client.close()

        # check if user listed in output
        cmd = self.master.run_command(['semanage', 'login', '-l'])
        assert testuser in cmd.stdout_text

        # call ipa user-del
        self.master.run_command(['ipa', 'user-del', testuser])

    def test_stageuser_show_as_alternate_admin(self):
        """
        Test that a user member of admins group can call stageuser-show
        and read the 'Kerberos Keys available' information.

        This is a test case for issue 7342
        """
        # kinit admin
        tasks.kinit_admin(self.master)

        # Create a new stage user 'stageuser' with a password
        stageuser = 'stageuser'
        password = 'Secret123'
        stageuser_password_confirmation = "%s\n%s\n" % (password,
                                                        password)
        self.master.run_command(['ipa', 'stageuser-add', stageuser,
                                 '--first', stageuser,
                                 '--last', stageuser,
                                 '--password'],
                                stdin_text=stageuser_password_confirmation)

        # kinit as altadmin
        self.master.run_command(['kinit', self.altadmin],
                                stdin_text=self.master.config.admin_password)

        # call ipa stageuser-show
        # the field Kerberos Keys available must contain True
        result = self.master.run_command(['ipa', 'stageuser-show', stageuser])
        assert 'Kerberos keys available: True' in result.stdout_text

    def test_user_add_withradius(self):
        """
        Test that a user with User Administrator role can call
        ipa user-add --radius myradius
        to create a user with an assigned Radius Proxy Server.

        This is a test case for issue 7570
        """
        # kinit admin
        tasks.kinit_admin(self.master)

        # Create a radius proxy server
        radiusproxy = 'myradius'
        secret = 'Secret123'
        radius_secret_confirmation = "%s\n%s\n" % (secret, secret)
        self.master.run_command(
            ['ipa', 'radiusproxy-add', radiusproxy,
             '--server', 'radius.example.com', '--secret'],
            stdin_text=radius_secret_confirmation)

        # Create a user with 'User Administrator' role
        altuser = 'specialuser'
        password = 'SpecialUser123'
        password_confirmation = "%s\n%s\n" % (password, password)
        self.master.run_command(
            ['ipa', 'user-add', altuser, '--first', altuser, '--last', altuser,
             '--password'],
            stdin_text=password_confirmation)
        self.master.run_command(
            ['ipa', 'role-add-member', "User Administrator",
             '--user', altuser])

        # kinit as altuser to initialize the password
        altuser_kinit = "%s\n%s\n%s\n" % (password, password, password)
        self.master.run_command(['kinit', altuser], stdin_text=altuser_kinit)
        # call ipa user-add with --radius=...
        # this call requires read access to radius proxy servers
        self.master.run_command(
            ['ipa', 'user-add', '--first', 'test', '--last', 'test',
             '--user-auth-type', 'radius', '--radius-username', 'testradius',
             'testradius', '--radius', radiusproxy])



class TestInstallClientNoAdmin(IntegrationTest):
    num_clients = 1

    def test_installclient_as_user_admin(self):
        """ipa-client-install should not use hardcoded admin for principal

        In ipaclient-install.log it should use the username that was entered
        earlier in the install process at the prompt.
        Related to : https://pagure.io/freeipa/issue/5406
        """
        client = self.clients[0]
        tasks.install_master(self.master)
        tasks.kinit_admin(self.master)
        username = 'testuser1'
        password = 'userSecretPassword123'
        password_confirmation = "%s\n%s\n" % (password,
                                              password)

        self.master.run_command(['ipa', 'user-add', username,
                                 '--first', username,
                                 '--last', username,
                                 '--password'],
                                stdin_text=password_confirmation)

        role_add = ['ipa', 'role-add', 'useradmin']
        self.master.run_command(role_add)
        self.master.run_command(['ipa', 'privilege-add', 'Add Hosts'])
        self.master.run_command(['ipa', 'privilege-add-permission',
                                 '--permissions', 'System: Add Hosts',
                                 'Add Hosts'])

        self.master.run_command(['ipa', 'role-add-privilege', 'useradmin',
                                 '--privileges', 'Host Enrollment'])

        self.master.run_command(['ipa', 'role-add-privilege', 'useradmin',
                                 '--privileges', 'Add Hosts'])

        role_member_add = ['ipa', 'role-add-member', 'useradmin',
                           '--users={}'.format(username)]
        self.master.run_command(role_member_add)
        user_kinit = "%s\n%s\n%s\n" % (password, password, password)
        self.master.run_command(['kinit', username],
                                stdin_text=user_kinit)
        tasks.install_client(
            self.master, client,
            extra_args=['--request-cert'],
            user=username, password=password
        )
        msg = "args=['/usr/bin/getent', 'passwd', '%s@%s']" % \
              (username, client.domain.name)
        install_log = client.get_file_contents(paths.IPACLIENT_INSTALL_LOG,
                                               encoding='utf-8')
        assert msg in install_log

        # check that user is able to request a host cert, too
        result = tasks.run_certutil(client, ['-L'], paths.IPA_NSSDB_DIR)
        assert 'Local IPA host' in result.stdout_text
        result = tasks.run_certutil(
            client,
            ['-K', '-f', paths.IPA_NSSDB_PWDFILE_TXT],
            paths.IPA_NSSDB_DIR
        )
        assert 'Local IPA host' in result.stdout_text



# Authors:
#   Jr Aquino <jr.aquino@citrixonline.com>
#
# Copyright (C) 2010  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ipalib import api
from ipalib import Str
from ipalib.plugable import Registry
from .baseldap import (
    LDAPObject,
    LDAPCreate,
    LDAPDelete,
    LDAPUpdate,
    LDAPSearch,
    LDAPRetrieve,
    LDAPAddMember,
    LDAPRemoveMember)
from ipalib import _, ngettext

__doc__ = _("""
Groups of Sudo Commands

Manage groups of Sudo Commands.

EXAMPLES:

 Add a new Sudo Command Group:
   ipa sudocmdgroup-add --desc='administrators commands' admincmds

 Remove a Sudo Command Group:
   ipa sudocmdgroup-del admincmds

 Manage Sudo Command Group membership, commands:
   ipa sudocmdgroup-add-member --sudocmds=/usr/bin/less --sudocmds=/usr/bin/vim admincmds

 Manage Sudo Command Group membership, commands:
   ipa sudocmdgroup-remove-member --sudocmds=/usr/bin/less admincmds

 Show a Sudo Command Group:
   ipa sudocmdgroup-show admincmds
""")

register = Registry()

topic = 'sudo'

@register()
class sudocmdgroup(LDAPObject):
    """
    Sudo Command Group object.
    """
    container_dn = api.env.container_sudocmdgroup
    object_name = _('sudo command group')
    object_name_plural = _('sudo command groups')
    object_class = ['ipaobject', 'ipasudocmdgrp']
    permission_filter_objectclasses = ['ipasudocmdgrp']
    default_attributes = [
        'cn', 'description', 'member',
    ]
    uuid_attribute = 'ipauniqueid'
    attribute_members = {
        'member': ['sudocmd'],
    }
    managed_permissions = {
        'System: Read Sudo Command Groups': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'businesscategory', 'cn', 'description', 'ipauniqueid',
                'member', 'o', 'objectclass', 'ou', 'owner', 'seealso',
                'memberuser', 'memberhost',
            },
        },
        'System: Add Sudo Command Group': {
            'ipapermright': {'add'},
            'replaces': [
                '(target = "ldap:///cn=*,cn=sudocmdgroups,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Add Sudo command group";allow (add) groupdn = "ldap:///cn=Add Sudo command group,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Sudo Administrator'},
        },
        'System: Delete Sudo Command Group': {
            'ipapermright': {'delete'},
            'replaces': [
                '(target = "ldap:///cn=*,cn=sudocmdgroups,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Delete Sudo command group";allow (delete) groupdn = "ldap:///cn=Delete Sudo command group,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Sudo Administrator'},
        },
        'System: Modify Sudo Command Group': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {'description'},
            'default_privileges': {'Sudo Administrator'},
        },
        'System: Manage Sudo Command Group Membership': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {'member'},
            'replaces': [
                '(targetattr = "member")(target = "ldap:///cn=*,cn=sudocmdgroups,cn=sudo,$SUFFIX")(version 3.0;acl "permission:Manage Sudo command group membership";allow (write) groupdn = "ldap:///cn=Manage Sudo command group membership,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Sudo Administrator'},
        },
    }

    label = _('Sudo Command Groups')
    label_singular = _('Sudo Command Group')

    takes_params = (
        Str('cn',
            cli_name='sudocmdgroup_name',
            label=_('Sudo Command Group'),
            primary_key=True,
            normalizer=lambda value: value.lower(),
        ),
        Str('description?',
            cli_name='desc',
            label=_('Description'),
            doc=_('Group description'),
        ),
        Str('membercmd_sudocmd?',
            label=_('Commands'),
            flags=['no_create', 'no_update', 'no_search'],
        ),
        Str('membercmd_sudocmdgroup?',
            label=_('Sudo Command Groups'),
            flags=['no_create', 'no_update', 'no_search'],
        ),
    )



@register()
class sudocmdgroup_add(LDAPCreate):
    __doc__ = _('Create new Sudo Command Group.')

    msg_summary = _('Added Sudo Command Group "%(value)s"')



@register()
class sudocmdgroup_del(LDAPDelete):
    __doc__ = _('Delete Sudo Command Group.')

    msg_summary = _('Deleted Sudo Command Group "%(value)s"')



@register()
class sudocmdgroup_mod(LDAPUpdate):
    __doc__ = _('Modify Sudo Command Group.')

    msg_summary = _('Modified Sudo Command Group "%(value)s"')



@register()
class sudocmdgroup_find(LDAPSearch):
    __doc__ = _('Search for Sudo Command Groups.')

    msg_summary = ngettext(
        '%(count)d Sudo Command Group matched',
        '%(count)d Sudo Command Groups matched', 0
    )



@register()
class sudocmdgroup_show(LDAPRetrieve):
    __doc__ = _('Display Sudo Command Group.')



@register()
class sudocmdgroup_add_member(LDAPAddMember):
    __doc__ = _('Add members to Sudo Command Group.')



@register()
class sudocmdgroup_remove_member(LDAPRemoveMember):
    __doc__ = _('Remove members from Sudo Command Group.')



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#
"""Distribution information

Known Linux distros with /etc/os-release
----------------------------------------

- alpine
- centos (like rhel, fedora)
- debian
- fedora
- rhel
- ubuntu (like debian)
"""
from __future__ import absolute_import

import importlib
import io
import re
import sys
import warnings

import six

import ipaplatform
try:
    from ipaplatform.override import OVERRIDE
except ImportError:
    OVERRIDE = None


# pylint: disable=no-name-in-module, import-error
if six.PY3:
    from collections.abc import Mapping
else:
    from collections import Mapping
# pylint: enable=no-name-in-module, import-error

_osrelease_line = re.compile(
    u"^(?!#)(?P<name>[a-zA-Z0-9_]+)="
    u"(?P<quote>[\"\']?)(?P<value>.+)(?P=quote)$"
)


def _parse_osrelease(filename='/etc/os-release'):
    """Parser for /etc/os-release for Linux distributions

    https://www.freedesktop.org/software/systemd/man/os-release.html
    """
    release = {}
    with io.open(filename, encoding='utf-8') as f:
        for line in f:
            mo = _osrelease_line.match(line)
            if mo is not None:
                release[mo.group('name')] = mo.group('value')
    if 'ID_LIKE' in release:
        release['ID_LIKE'] = tuple(
            v.strip()
            for v in release['ID_LIKE'].split(' ')
            if v.strip()
        )
    else:
        release["ID_LIKE"] = ()
    # defaults
    release.setdefault('NAME', 'Linux')
    release.setdefault('ID', 'linux')
    release.setdefault('VERSION', '')
    release.setdefault('VERSION_ID', '')
    return release


class OSInfo(Mapping):
    __slots__ = ('_info', '_platform')

    bsd_family = (
        'freebsd',
        'openbsd',
        'netbsd',
        'dragonfly',
        'gnukfreebsd'
    )

    def __init__(self):
        if sys.platform.startswith('linux'):
            # Linux, get distribution from /etc/os-release
            info = self._handle_linux()
        elif sys.platform == 'win32':
            info = self._handle_win32()
        elif sys.platform == 'darwin':
            info = self._handle_darwin()
        elif sys.platform.startswith(self.bsd_family):
            info = self._handle_bsd()
        else:
            raise ValueError("Unsupported platform: {}".format(sys.platform))
        self._info = info
        self._platform = None

    def _handle_linux(self):
        """Detect Linux distribution from /etc/os-release
        """
        try:
            return _parse_osrelease()
        except Exception as e:
            warnings.warn("Failed to read /etc/os-release: {}".format(e))
            return {
                'NAME': 'Linux',
                'ID': 'linux',
            }

    def _handle_win32(self):
        """Windows 32 or 64bit platform
        """
        return {
            'NAME': 'Windows',
            'ID': 'win32',
        }

    def _handle_darwin(self):
        """Handle macOS / Darwin platform
        """
        return {
            'NAME': 'macOS',
            'ID': 'macos',
        }

    def _handle_bsd(self):
        """Handle BSD-like platforms
        """
        platform = sys.platform
        simple = platform.rstrip('0123456789')
        id_like = []
        if simple != platform:
            id_like.append(simple)
        return {
            'NAME': platform,
            'ID': platform,
            'ID_LIKE': tuple(id_like),
        }

    def __getitem__(self, item):
        return self._info[item]

    def __iter__(self):
        return iter(self._info)

    def __len__(self):
        return len(self._info)

    @property
    def name(self):
        """OS name (user)
        """
        return self._info['NAME']

    @property
    def id(self):
        """Lower case OS identifier
        """
        return self._info['ID']

    @property
    def id_like(self):
        """Related / similar OS
        """
        return self._info.get('ID_LIKE', ())

    @property
    def version(self):
        """Version number and name of OS (for user)
        """
        return self._info.get('VERSION')

    @property
    def version_number(self):
        """Version number tuple based on version_id
        """
        version_id = self._info.get('VERSION_ID')
        if not version_id:
            return ()
        return tuple(int(p) for p in version_id.split('.'))

    @property
    def platform_ids(self):
        """Ordered tuple of detected platforms (including override)
        """
        platforms = []
        if OVERRIDE is not None:
            # allow RPM and Debian packages to override platform
            platforms.append(OVERRIDE)
        if OVERRIDE != self.id:
            platforms.append(self.id)
        platforms.extend(self.id_like)
        return tuple(platforms)

    @property
    def platform(self):
        if self._platform is not None:
            return self._platform
        for platform in self.platform_ids:
            try:
                importlib.import_module('ipaplatform.{}'.format(platform))
            except ImportError:
                pass
            else:
                self._platform = platform
                return platform
        raise ImportError('No ipaplatform available for "{}"'.format(
                          ', '.join(self.platform_ids)))


osinfo = OSInfo()
ipaplatform.NAME = osinfo.platform

if __name__ == '__main__':
    import pprint
    pprint.pprint(dict(osinfo))



#!/usr/bin/python3
"""Helper script to test LGTM config

$ contrib/lgtm_container.py > Dockerfile
$ docker build -t lgtm .
"""
import os
import yaml

LGTM_YML = os.path.join(os.path.dirname(__file__), '..', '.lgtm.yml')


def main():
    with open(LGTM_YML) as f:
        cfg = yaml.safe_load(f)

    python = cfg['extraction']['python']

    print("""\
    FROM ubuntu:bionic
    RUN apt-get update && \
        apt-get install -y {dpkg} python3-venv && \
        apt-get clean
    RUN python3 -m venv /venv
    RUN /venv/bin/pip install wheel
    RUN /venv/bin/pip install {pypkg}
    ADD . /freeipa
    RUN cd /freeipa && ./autogen.sh --with-ipaplatform=debian
    """.format(
        dpkg=' '.join(python['prepare']['packages']),
        pypkg=' '.join(python['python_setup']['requirements'])
    ))


if __name__ == '__main__':
    main()



# Authors:
#   Rob Crittenden <rcritten@redhat.com>
#   Pavel Zuna <pzuna@redhat.com>
#
# Copyright (C) 2009  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import six

import logging

from ipalib import api
from ipalib import Int, Str, Flag
from ipalib.constants import PATTERN_GROUPUSER_NAME
from ipalib.plugable import Registry
from .baseldap import (
    add_external_post_callback,
    pkey_to_value,
    remove_external_post_callback,
    LDAPObject,
    LDAPCreate,
    LDAPUpdate,
    LDAPDelete,
    LDAPSearch,
    LDAPRetrieve,
    LDAPAddMember,
    LDAPRemoveMember,
    LDAPQuery,
)
from .idviews import remove_ipaobject_overrides
from . import baseldap
from ipalib import _, ngettext
from ipalib import errors
from ipalib import output
from ipapython.dn import DN

if six.PY3:
    unicode = str

logger = logging.getLogger(__name__)

if api.env.in_server and api.env.context in ['lite', 'server']:
    try:
        import ipaserver.dcerpc
        _dcerpc_bindings_installed = True
    except ImportError:
        _dcerpc_bindings_installed = False

__doc__ = _("""
Groups of users

Manage groups of users, groups, or services. By default, new groups are POSIX
groups. You can add the --nonposix option to the group-add command to mark a
new group as non-POSIX. You can use the --posix argument with the group-mod
command to convert a non-POSIX group into a POSIX group. POSIX groups cannot be
converted to non-POSIX groups.

Every group must have a description.

POSIX groups must have a Group ID (GID) number. Changing a GID is
supported but can have an impact on your file permissions. It is not necessary
to supply a GID when creating a group. IPA will generate one automatically
if it is not provided.

Groups members can be users, other groups, and Kerberos services. In POSIX
environments only users will be visible as group members, but nested groups and
groups of services can be used for IPA management purposes.

EXAMPLES:

 Add a new group:
   ipa group-add --desc='local administrators' localadmins

 Add a new non-POSIX group:
   ipa group-add --nonposix --desc='remote administrators' remoteadmins

 Convert a non-POSIX group to posix:
   ipa group-mod --posix remoteadmins

 Add a new POSIX group with a specific Group ID number:
   ipa group-add --gid=500 --desc='unix admins' unixadmins

 Add a new POSIX group and let IPA assign a Group ID number:
   ipa group-add --desc='printer admins' printeradmins

 Remove a group:
   ipa group-del unixadmins

 To add the "remoteadmins" group to the "localadmins" group:
   ipa group-add-member --groups=remoteadmins localadmins

 Add multiple users to the "localadmins" group:
   ipa group-add-member --users=test1 --users=test2 localadmins

 To add Kerberos services to the "printer admins" group:
   ipa group-add-member --services=CUPS/some.host printeradmins

 Remove a user from the "localadmins" group:
   ipa group-remove-member --users=test2 localadmins

 Display information about a named group.
   ipa group-show localadmins

External group membership is designed to allow users from trusted domains
to be mapped to local POSIX groups in order to actually use IPA resources.
External members should be added to groups that specifically created as
external and non-POSIX. Such group later should be included into one of POSIX
groups.

An external group member is currently a Security Identifier (SID) as defined by
the trusted domain. When adding external group members, it is possible to
specify them in either SID, or DOM\\name, or name@domain format. IPA will attempt
to resolve passed name to SID with the use of Global Catalog of the trusted domain.

Example:

1. Create group for the trusted domain admins' mapping and their local POSIX group:

   ipa group-add --desc='<ad.domain> admins external map' ad_admins_external --external
   ipa group-add --desc='<ad.domain> admins' ad_admins

2. Add security identifier of Domain Admins of the <ad.domain> to the ad_admins_external
   group:

   ipa group-add-member ad_admins_external --external 'AD\\Domain Admins'

3. Allow members of ad_admins_external group to be associated with ad_admins POSIX group:

   ipa group-add-member ad_admins --groups ad_admins_external

4. List members of external members of ad_admins_external group to see their SIDs:

   ipa group-show ad_admins_external
""")

register = Registry()

PROTECTED_GROUPS = (u'admins', u'trust admins', u'default smb group')


ipaexternalmember_param = Str('ipaexternalmember*',
            cli_name='external',
            label=_('External member'),
            doc=_('Members of a trusted domain in DOM\\name or name@domain form'),
            flags=['no_create', 'no_update', 'no_search'],
        )


@register()
class group(LDAPObject):
    """
    Group object.
    """
    container_dn = api.env.container_group
    object_name = _('group')
    object_name_plural = _('groups')
    object_class = ['ipausergroup']
    object_class_config = 'ipagroupobjectclasses'
    possible_objectclasses = ['posixGroup', 'mepManagedEntry', 'ipaExternalGroup']
    permission_filter_objectclasses = ['posixgroup', 'ipausergroup']
    search_attributes_config = 'ipagroupsearchfields'
    default_attributes = [
        'cn', 'description', 'gidnumber', 'member', 'memberof',
        'memberindirect', 'memberofindirect', 'ipaexternalmember',
    ]
    uuid_attribute = 'ipauniqueid'
    attribute_members = {
        'member': ['user', 'group', 'service'],
        'memberof': ['group', 'netgroup', 'role', 'hbacrule', 'sudorule'],
        'memberindirect': ['user', 'group', 'service'],
        'memberofindirect': ['group', 'netgroup', 'role', 'hbacrule',
        'sudorule'],
    }
    allow_rename = True
    managed_permissions = {
        'System: Read Groups': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'anonymous',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'businesscategory', 'cn', 'description', 'gidnumber',
                'ipaexternalmember', 'ipauniqueid', 'mepmanagedby', 'o',
                'objectclass', 'ou', 'owner', 'seealso',
                'ipantsecurityidentifier'
            },
        },
        'System: Read Group Membership': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'member', 'memberof', 'memberuid', 'memberuser', 'memberhost',
            },
        },
        'System: Read External Group Membership': {
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'ipaexternalmember',
            },
        },
        'System: Add Groups': {
            'ipapermright': {'add'},
            'replaces': [
                '(target = "ldap:///cn=*,cn=groups,cn=accounts,$SUFFIX")(version 3.0;acl "permission:Add Groups";allow (add) groupdn = "ldap:///cn=Add Groups,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Group Administrators'},
        },
        'System: Modify Group Membership': {
            'ipapermright': {'write'},
            'ipapermtargetfilter': [
                '(objectclass=ipausergroup)',
                '(!(cn=admins))',
            ],
            'ipapermdefaultattr': {'member'},
            'replaces': [
                '(targetattr = "member")(target = "ldap:///cn=*,cn=groups,cn=accounts,$SUFFIX")(version 3.0;acl "permission:Modify Group membership";allow (write) groupdn = "ldap:///cn=Modify Group membership,cn=permissions,cn=pbac,$SUFFIX";)',
                '(targetfilter = "(!(cn=admins))")(targetattr = "member")(target = "ldap:///cn=*,cn=groups,cn=accounts,$SUFFIX")(version 3.0;acl "permission:Modify Group membership";allow (write) groupdn = "ldap:///cn=Modify Group membership,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {
                'Group Administrators', 'Modify Group membership'
            },
        },
        'System: Modify External Group Membership': {
            'ipapermright': {'write'},
            'ipapermtargetfilter': [
                '(objectclass=ipaexternalgroup)',
            ],
            'ipapermdefaultattr': {'ipaexternalmember'},
            'default_privileges': {
                'Group Administrators', 'Modify Group membership'
            },
        },
        'System: Modify Groups': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {
                'cn', 'description', 'gidnumber', 'ipauniqueid',
                'mepmanagedby', 'objectclass'
            },
            'replaces': [
                '(targetattr = "cn || description || gidnumber || objectclass || mepmanagedby || ipauniqueid")(target = "ldap:///cn=*,cn=groups,cn=accounts,$SUFFIX")(version 3.0;acl "permission:Modify Groups";allow (write) groupdn = "ldap:///cn=Modify Groups,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Group Administrators'},
        },
        'System: Remove Groups': {
            'ipapermright': {'delete'},
            'replaces': [
                '(target = "ldap:///cn=*,cn=groups,cn=accounts,$SUFFIX")(version 3.0;acl "permission:Remove Groups";allow (delete) groupdn = "ldap:///cn=Remove Groups,cn=permissions,cn=pbac,$SUFFIX";)',
            ],
            'default_privileges': {'Group Administrators'},
        },
        'System: Read Group Compat Tree': {
            'non_object': True,
            'ipapermbindruletype': 'anonymous',
            'ipapermlocation': api.env.basedn,
            'ipapermtarget': DN('cn=groups', 'cn=compat', api.env.basedn),
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'objectclass', 'cn', 'memberuid', 'gidnumber',
            },
        },
        'System: Read Group Views Compat Tree': {
            'non_object': True,
            'ipapermbindruletype': 'anonymous',
            'ipapermlocation': api.env.basedn,
            'ipapermtarget': DN('cn=groups', 'cn=*', 'cn=views', 'cn=compat', api.env.basedn),
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'objectclass', 'cn', 'memberuid', 'gidnumber',
            },
        },
    }

    label = _('User Groups')
    label_singular = _('User Group')

    takes_params = (
        Str('cn',
            pattern=PATTERN_GROUPUSER_NAME,
            pattern_errmsg='may only include letters, numbers, _, -, . and $',
            maxlength=255,
            cli_name='group_name',
            label=_('Group name'),
            primary_key=True,
            normalizer=lambda value: value.lower(),
        ),
        Str('description?',
            cli_name='desc',
            label=_('Description'),
            doc=_('Group description'),
        ),
        Int('gidnumber?',
            cli_name='gid',
            label=_('GID'),
            doc=_('GID (use this option to set it manually)'),
            minvalue=1,
        ),
        ipaexternalmember_param,
    )


@register()
class group_add(LDAPCreate):
    __doc__ = _('Create a new group.')

    msg_summary = _('Added group "%(value)s"')

    takes_options = LDAPCreate.takes_options + (
        Flag('nonposix',
             cli_name='nonposix',
             doc=_('Create as a non-POSIX group'),
             default=False,
        ),
        Flag('external',
             cli_name='external',
             doc=_('Allow adding external non-IPA members from trusted domains'),
             default=False,
        ),
    )

    def pre_callback(self, ldap, dn, entry_attrs, attrs_list, *keys, **options):
        # As both 'external' and 'nonposix' options have default= set for
        # them, they will always be present in options dict, thus we can
        # safely reference the values
        assert isinstance(dn, DN)
        if options['external']:
            entry_attrs['objectclass'].append('ipaexternalgroup')
            if 'gidnumber' in options:
                raise errors.MutuallyExclusiveError(reason=_('gid cannot be set for external group'))
        elif not options['nonposix']:
            entry_attrs['objectclass'].append('posixgroup')
            if 'gidnumber' not in options:
                entry_attrs['gidnumber'] = baseldap.DNA_MAGIC
        return dn


@register()
class group_del(LDAPDelete):
    __doc__ = _('Delete group.')

    msg_summary = _('Deleted group "%(value)s"')

    def pre_callback(self, ldap, dn, *keys, **options):
        assert isinstance(dn, DN)
        config = ldap.get_ipa_config()
        def_primary_group = config.get('ipadefaultprimarygroup', '')
        def_primary_group_dn = self.obj.get_dn(def_primary_group)
        if dn == def_primary_group_dn:
            raise errors.DefaultGroupError()
        group_attrs = self.obj.methods.show(
            self.obj.get_primary_key_from_dn(dn), all=True
        )['result']
        if keys[0] in PROTECTED_GROUPS:
            raise errors.ProtectedEntryError(label=_(u'group'), key=keys[0],
                reason=_(u'privileged group'))
        if 'mepmanagedby' in group_attrs:
            raise errors.ManagedGroupError()

        # Remove any ID overrides tied with this group
        remove_ipaobject_overrides(ldap, self.obj.api, dn)

        return dn

    def post_callback(self, ldap, dn, *keys, **options):
        assert isinstance(dn, DN)
        try:
            # A user removing a group may have no rights to remove
            # an associated policy. Make sure we log an explanation
            # in the Apache logs for this.
            api.Command['pwpolicy_del'](keys[-1])
        except errors.ACIError:
            logger.warning(
                "While removing group %s, user lacked permissions "
                "to remove corresponding password policy. This is "
                "not an issue and can be ignored.", keys[-1]
            )
        except errors.NotFound:
            pass

        return True


@register()
class group_mod(LDAPUpdate):
    __doc__ = _('Modify a group.')

    msg_summary = _('Modified group "%(value)s"')

    takes_options = LDAPUpdate.takes_options + (
        Flag('posix',
             cli_name='posix',
             doc=_('change to a POSIX group'),
        ),
        Flag('external',
             cli_name='external',
             doc=_('change to support external non-IPA members from trusted domains'),
             default=False,
        ),
    )

    def pre_callback(self, ldap, dn, entry_attrs, *keys, **options):
        assert isinstance(dn, DN)

        is_protected_group = keys[-1] in PROTECTED_GROUPS

        if 'rename' in options or 'cn' in entry_attrs:
            if is_protected_group:
                raise errors.ProtectedEntryError(label=u'group', key=keys[-1],
                    reason=u'Cannot be renamed')

        if ('posix' in options and options['posix']) or 'gidnumber' in options:
            old_entry_attrs = ldap.get_entry(dn, ['objectclass'])
            dn = old_entry_attrs.dn
            if 'ipaexternalgroup' in old_entry_attrs['objectclass']:
                raise errors.ExternalGroupViolation()
            if 'posixgroup' in old_entry_attrs['objectclass']:
                if options['posix']:
                    raise errors.AlreadyPosixGroup()
            else:
                old_entry_attrs['objectclass'].append('posixgroup')
                entry_attrs['objectclass'] = old_entry_attrs['objectclass']
                if 'gidnumber' not in options:
                    entry_attrs['gidnumber'] = baseldap.DNA_MAGIC

        if options['external']:
            if is_protected_group:
                raise errors.ProtectedEntryError(label=u'group', key=keys[-1],
                    reason=u'Cannot support external non-IPA members')
            old_entry_attrs = ldap.get_entry(dn, ['objectclass'])
            dn = old_entry_attrs.dn
            if 'posixgroup' in old_entry_attrs['objectclass']:
                raise errors.PosixGroupViolation()
            if 'ipaexternalgroup' in old_entry_attrs['objectclass']:
                raise errors.AlreadyExternalGroup()
            else:
                old_entry_attrs['objectclass'].append('ipaexternalgroup')
                entry_attrs['objectclass'] = old_entry_attrs['objectclass']

        # Can't check for this in a validator because we lack context
        if 'gidnumber' in options and options['gidnumber'] is None:
            raise errors.RequirementError(name='gidnumber')
        return dn

    def exc_callback(self, keys, options, exc, call_func, *call_args, **call_kwargs):
        # Check again for GID requirement in case someone tried to clear it
        # using --setattr.
        if call_func.__name__ == 'update_entry':
            if isinstance(exc, errors.ObjectclassViolation):
                if 'gidNumber' in str(exc) and 'posixGroup' in str(exc):
                    raise errors.RequirementError(name='gidnumber')
        raise exc


@register()
class group_find(LDAPSearch):
    __doc__ = _('Search for groups.')

    member_attributes = ['member', 'memberof']

    msg_summary = ngettext(
        '%(count)d group matched', '%(count)d groups matched', 0
    )

    takes_options = LDAPSearch.takes_options + (
        Flag('private',
            cli_name='private',
            doc=_('search for private groups'),
        ),
        Flag('posix',
             cli_name='posix',
             doc=_('search for POSIX groups'),
        ),
        Flag('external',
             cli_name='external',
             doc=_('search for groups with support of external non-IPA members from trusted domains'),
        ),
        Flag('nonposix',
             cli_name='nonposix',
             doc=_('search for non-POSIX groups'),
        ),
    )

    def pre_callback(self, ldap, filter, attrs_list, base_dn, scope,
                     criteria=None, **options):
        assert isinstance(base_dn, DN)

        # filter groups by pseudo type
        filters = []
        if options['posix']:
            search_kw = {'objectclass': ['posixGroup']}
            filters.append(ldap.make_filter(search_kw, rules=ldap.MATCH_ALL))
        if options['external']:
            search_kw = {'objectclass': ['ipaExternalGroup']}
            filters.append(ldap.make_filter(search_kw, rules=ldap.MATCH_ALL))
        if options['nonposix']:
            search_kw = {'objectclass': ['posixGroup' , 'ipaExternalGroup']}
            filters.append(ldap.make_filter(search_kw, rules=ldap.MATCH_NONE))

        # if looking for private groups, we need to create a new search filter,
        # because private groups have different object classes
        if options['private']:
            # filter based on options, oflt
            search_kw = self.args_options_2_entry(**options)
            search_kw['objectclass'] = ['posixGroup', 'mepManagedEntry']
            oflt = ldap.make_filter(search_kw, rules=ldap.MATCH_ALL)

            # filter based on 'criteria' argument
            search_kw = {}
            config = ldap.get_ipa_config()
            attrs = config.get(self.obj.search_attributes_config, [])
            if len(attrs) == 1 and isinstance(attrs[0], str):
                search_attrs = attrs[0].split(',')
                for a in search_attrs:
                    search_kw[a] = criteria
            cflt = ldap.make_filter(search_kw, exact=False)

            filter = ldap.combine_filters((oflt, cflt), rules=ldap.MATCH_ALL)
        elif filters:
            filters.append(filter)
            filter = ldap.combine_filters(filters, rules=ldap.MATCH_ALL)
        return (filter, base_dn, scope)


@register()
class group_show(LDAPRetrieve):
    __doc__ = _('Display information about a named group.')

    def post_callback(self, ldap, dn, entry_attrs, *keys, **options):
        assert isinstance(dn, DN)
        if ('ipaexternalmember' in entry_attrs and
            len(entry_attrs['ipaexternalmember']) > 0 and
            'trust_resolve' in self.Command and
            not options.get('raw', False)):
            sids = entry_attrs['ipaexternalmember']
            result = self.Command.trust_resolve(sids=sids)
            for entry in result['result']:
                try:
                    idx = sids.index(entry['sid'][0])
                    sids[idx] = entry['name'][0]
                except ValueError:
                    pass
        return dn


@register()
class group_add_member(LDAPAddMember):
    __doc__ = _('Add members to a group.')

    takes_options = (ipaexternalmember_param,)

    def post_callback(self, ldap, completed, failed, dn, entry_attrs, *keys, **options):
        assert isinstance(dn, DN)
        result = (completed, dn)
        if 'ipaexternalmember' in options:
            if not _dcerpc_bindings_installed:
                raise errors.NotFound(reason=_('Cannot perform external member validation without '
                                      'Samba 4 support installed. Make sure you have installed '
                                      'server-trust-ad sub-package of IPA on the server'))
            domain_validator = ipaserver.dcerpc.DomainValidator(self.api)
            if not domain_validator.is_configured():
                raise errors.NotFound(reason=_('Cannot perform join operation without own domain configured. '
                                      'Make sure you have run ipa-adtrust-install on the IPA server first'))
            sids = []
            failed_sids = []
            for sid in options['ipaexternalmember']:
                if domain_validator.is_trusted_sid_valid(sid):
                    sids.append(sid)
                else:
                    try:
                        actual_sid = domain_validator.get_trusted_domain_object_sid(sid)
                    except errors.PublicError as e:
                        failed_sids.append((sid, e.strerror))
                    else:
                        sids.append(actual_sid)
            restore = []
            if 'member' in failed and 'group' in failed['member']:
                restore = failed['member']['group']
            failed['member']['group'] = list((id, id) for id in sids)
            result = add_external_post_callback(ldap, dn, entry_attrs,
                                                failed=failed,
                                                completed=completed,
                                                memberattr='member',
                                                membertype='group',
                                                externalattr='ipaexternalmember',
                                                normalize=False)
            failed['member']['group'] += restore + failed_sids
        return result


@register()
class group_remove_member(LDAPRemoveMember):
    __doc__ = _('Remove members from a group.')

    takes_options = (ipaexternalmember_param,)

    def pre_callback(self, ldap, dn, found, not_found, *keys, **options):
        assert isinstance(dn, DN)
        if keys[0] in PROTECTED_GROUPS and 'user' in options:
            protected_group_name = keys[0]
            result = api.Command.group_show(protected_group_name)
            users_left = set(result['result'].get('member_user', []))
            users_deleted = set(options['user'])
            if users_left.issubset(users_deleted):
                raise errors.LastMemberError(key=sorted(users_deleted)[0],
                    label=_(u'group'), container=protected_group_name)
        return dn

    def post_callback(self, ldap, completed, failed, dn, entry_attrs, *keys, **options):
        assert isinstance(dn, DN)
        result = (completed, dn)
        if 'ipaexternalmember' in options:
            if not _dcerpc_bindings_installed:
                raise errors.NotFound(reason=_('Cannot perform external member validation without '
                                               'Samba 4 support installed. Make sure you have installed '
                                               'server-trust-ad sub-package of IPA on the server'))
            domain_validator = ipaserver.dcerpc.DomainValidator(self.api)
            if not domain_validator.is_configured():
                raise errors.NotFound(reason=_('Cannot perform join operation without own domain configured. '
                                               'Make sure you have run ipa-adtrust-install on the IPA server first'))
            sids = []
            failed_sids = []
            for sid in options['ipaexternalmember']:
                if domain_validator.is_trusted_sid_valid(sid):
                    sids.append(sid)
                else:
                    try:
                        actual_sid = domain_validator.get_trusted_domain_object_sid(sid)
                    except errors.PublicError as e:
                        failed_sids.append((sid, unicode(e)))
                    else:
                        sids.append(actual_sid)
            restore = []
            if 'member' in failed and 'group' in failed['member']:
                restore = failed['member']['group']
            failed['member']['group'] = list((id, id) for id in sids)
            result = remove_external_post_callback(ldap, dn, entry_attrs,
                                                failed=failed,
                                                completed=completed,
                                                memberattr='member',
                                                membertype='group',
                                                externalattr='ipaexternalmember',
                                                )
            failed['member']['group'] += restore + failed_sids
        return result


@register()
class group_detach(LDAPQuery):
    __doc__ = _('Detach a managed group from a user.')

    has_output = output.standard_value
    msg_summary = _('Detached group "%(value)s" from user "%(value)s"')

    def execute(self, *keys, **options):
        """
        This requires updating both the user and the group. We first need to
        verify that both the user and group can be updated, then we go
        about our work. We don't want a situation where only the user or
        group can be modified and we're left in a bad state.
        """
        ldap = self.obj.backend

        group_dn = self.obj.get_dn(*keys, **options)
        user_dn = self.api.Object['user'].get_dn(*keys)

        try:
            user_attrs = ldap.get_entry(user_dn)
        except errors.NotFound:
            raise self.obj.handle_not_found(*keys)
        is_managed = self.obj.has_objectclass(
            user_attrs['objectclass'], 'mepmanagedentry'
        )
        if (not ldap.can_write(user_dn, "objectclass") or
                not ldap.can_write(user_dn, "mepManagedEntry")
                and is_managed):
            raise errors.ACIError(
                info=_('not allowed to modify user entries')
            )

        group_attrs = ldap.get_entry(group_dn)
        is_managed = self.obj.has_objectclass(
            group_attrs['objectclass'], 'mepmanagedby'
        )
        if (not ldap.can_write(group_dn, "objectclass") or
                not ldap.can_write(group_dn, "mepManagedBy")
                and is_managed):
            raise errors.ACIError(
                info=_('not allowed to modify group entries')
            )

        objectclasses = user_attrs['objectclass']
        try:
            i = objectclasses.index('mepOriginEntry')
            del objectclasses[i]
            user_attrs['mepManagedEntry'] = None
            ldap.update_entry(user_attrs)
        except ValueError:
            # Somehow the user isn't managed, let it pass for now. We'll
            # let the group throw "Not managed".
            pass

        group_attrs = ldap.get_entry(group_dn)
        objectclasses = group_attrs['objectclass']
        try:
            i = objectclasses.index('mepManagedEntry')
        except ValueError:
            # this should never happen
            raise errors.NotFound(reason=_('Not a managed group'))
        del objectclasses[i]

        # Make sure the resulting group has the default group objectclasses
        config = ldap.get_ipa_config()
        def_objectclass = config.get(
            self.obj.object_class_config, objectclasses
        )
        objectclasses = list(set(def_objectclass + objectclasses))

        group_attrs['mepManagedBy'] = None
        group_attrs['objectclass'] = objectclasses
        ldap.update_entry(group_attrs)

        return dict(
            result=True,
            value=pkey_to_value(keys[0], options),
        )
